

__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque",
               "Holden Parks", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


from functools import partial
from PySide import QtCore, QtGui
from PySide.QtUiTools import QUiLoader
from psutil import cpu_count
import pyqtgraph as pg
from pyqtgraph import parametertree as pt
from collections import deque
import xicam.widgets.featurewidgets as fw
from xicam.widgets.customwidgets import DataTreeWidget
import reconpkg
import os
import viewers


class UIform(object):
    """
    Class for tomography plugin ui setup.

    Attributes
    ----------

    toolbar : QtGui.QToolBar
        Toolbar shown in plugin
    leftmodes : list of tuples
        leftmodes list for standard base plugin initialization [widget, tab icon]
    righmodes : list of tuples
        rightmodes list for standard base plugin initialization [widget, tab icon]
    param_form : QtGui.QStackedWidget
        Container for ParameterTree's used to display function parameters
    functionWidget : QtGui.QWidget
        Workflow pipeline GUI in plugin's leftmodes
    property_table : pyqtgraph.TableWidget
        TableWidget to display dataset metadata on right widget
    config_params : pyqtgraph.Parameter
        Parameter holding the run configuration parameters (ie sino start, sino end, sino step, number of cores)

    Methods
    -------
    connectTriggers
        Connect leftwidget (function mangement buttons) triggers to corresponding slots
    setConfigParams
        Sets configuration parameters in pg.Parameter inside rightwidget
    """

    def setupUi(self):
        """Set up the UI for tomography plugin"""

        self.toolbar = Toolbar()
        self.centerwidget = QtGui.QTabWidget()
        self.centerwidget.setDocumentMode(True)
        self.centerwidget.setTabsClosable(True)
        self.bottomwidget = viewers.RunConsole()
        self.functionwidget = QUiLoader().load('xicam/gui/tomographyleft.ui')
        self.functionwidget.functionsList.setAlignment(QtCore.Qt.AlignBottom)

        self.functionwidget.addFunctionButton.setToolTip('Add function to pipeline')
        self.functionwidget.clearButton.setToolTip('Clear pipeline')
        self.functionwidget.fileButton.setToolTip('Save/Load pipeline')
        self.functionwidget.moveDownButton.setToolTip('Move selected function down')
        self.functionwidget.moveUpButton.setToolTip('Move selected function up')

        self.addfunctionmenu = QtGui.QMenu()
        self.functionwidget.addFunctionButton.setMenu(self.addfunctionmenu)
        self.functionwidget.addFunctionButton.setPopupMode(QtGui.QToolButton.ToolButtonPopupMode.InstantPopup)
        self.functionwidget.addFunctionButton.setArrowType(QtCore.Qt.NoArrow)

        filefuncmenu = QtGui.QMenu()
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_55.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.openaction = QtGui.QAction(icon, 'Open', filefuncmenu)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_59.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.saveaction = QtGui.QAction(icon, 'Save', filefuncmenu)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_56.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshaction = QtGui.QAction(icon, 'Reset', filefuncmenu)
        filefuncmenu.addActions([self.openaction, self.saveaction, self.refreshaction])

        self.functionwidget.fileButton.setMenu(filefuncmenu)
        self.functionwidget.fileButton.setPopupMode(QtGui.QToolButton.ToolButtonPopupMode.InstantPopup)
        self.functionwidget.fileButton.setArrowType(QtCore.Qt.NoArrow)

        leftwidget = QtGui.QSplitter(QtCore.Qt.Vertical)

        paramtree = pt.ParameterTree()
        self.param_form = QtGui.QStackedWidget()
        self.param_form.addWidget(paramtree)
        self.property_table = pg.TableWidget()
        self.property_table.verticalHeader().hide()
        self.property_table.horizontalHeader().setStretchLastSection(True)
        self.property_table.setSizePolicy(QtGui.QSizePolicy.Preferred, QtGui.QSizePolicy.Expanding)
        leftwidget.addWidget(self.param_form)
        leftwidget.addWidget(self.functionwidget)

        self.queue = ReconManager()

        logwidget = QtGui.QStackedWidget()
        logwidget.addWidget(self.bottomwidget)

        icon_functions = QtGui.QIcon(QtGui.QPixmap("xicam/gui/icons_49.png"))
        icon_properties = QtGui.QIcon(QtGui.QPixmap("xicam/gui/icons_61.png")) #metadata icon
        icon_queue = QtGui.QIcon(QtGui.QPixmap("xicam/gui/icons_63.png"))
        icon_log = QtGui.QIcon(QtGui.QPixmap("xicam/gui/icons_64.png"))
        self.leftmodes = [(leftwidget, icon_functions),(self.queue, icon_queue), (logwidget, icon_log),
                          (self.property_table, icon_properties)]


    def connectTriggers(self, open, save, reset, moveup, movedown, clear):
        """
        Connect leftwidget (function mangement buttons) triggers to corresponding slots

        Parameters
        ----------
        open : QtCore.Slot
            Slot to handle signal from open button
        save QtCore.Slot
            Slot to handle signal from save button
        reset QtCore.Slot
            Slot to handle signal from reset button
        moveup QtCore.Slot
            Slot to handle signal to move a function widget upwards
        movedown QtCore.Slot
            Slot to handle signal to move a function widget downwards
        clear QtCore.Slot
            Slot to handle signal from clear button
        """

        self.openaction.triggered.connect(open)
        self.saveaction.triggered.connect(save)
        self.refreshaction.triggered.connect(reset)
        self.functionwidget.moveDownButton.clicked.connect(moveup)
        self.functionwidget.moveUpButton.clicked.connect(movedown)
        self.functionwidget.clearButton.clicked.connect(clear)


def build_function_menu(menu, functree, functiondata, actionslot):
    """
    Builds the function menu's and submenu's anc connects them to the corresponding slot to add them to the workflow
    pipeline

    Parameters
    ----------
    menu : QtGui.QMenu
        Menu object to populate with submenu's and actions
    functree : dict
        Dictionary specifying the depth levels of functions. See functions.yml entry "Functions"
    functiondata : dict
        Dictionary with function information. See function_names.yml
    actionslot : QtCore.Slot
        slot where the function action triggered signal shoud be connected
    """

    for func, subfuncs in functree.iteritems():
        if len(subfuncs) > 1 or func != subfuncs[0]:
            funcmenu = QtGui.QMenu(func)
            menu.addMenu(funcmenu)
            for subfunc in subfuncs:
                if isinstance(subfuncs, dict) and len(subfuncs[subfunc]) > 0:
                    optsmenu = QtGui.QMenu(subfunc)
                    funcmenu.addMenu(optsmenu)
                    for opt in subfuncs[subfunc]:
                        funcaction = QtGui.QAction(opt, funcmenu)
                        try:
                            funcaction.triggered.connect(partial(actionslot, func, opt,
                                                                 reconpkg.packages[functiondata[opt][1]]))
                            optsmenu.addAction(funcaction)
                        except KeyError:
                            pass
                else:
                    funcaction = QtGui.QAction(subfunc, funcmenu)
                    try:
                        funcaction.triggered.connect(partial(actionslot, func, subfunc,
                                                             reconpkg.packages[functiondata[subfunc][1]]))
                        funcmenu.addAction(funcaction)
                    except KeyError:
                        pass
        elif len(subfuncs) == 1:
            try:
                funcaction = QtGui.QAction(func, menu)
                funcaction.triggered.connect(partial(actionslot, func, func, reconpkg.packages[functiondata[func][1]]))
                menu.addAction(funcaction)
            except KeyError:
                pass


class Toolbar(QtGui.QToolBar):
    """
    QToolbar subclass used in Tomography plugin

    Attributes
    ----------
    actionRun_SlicePreview : QtGui.QAction
    actionRun_3DPreview : QtGui.QAction
    actionRun_FullRecon : QtGui.QAction
    actionCenter : QtGui.QAction
    actionROI : QtGui.QAction
    actionPolyMask : QtGui.QAction
    actionCircMask : QtGui.QAction
    actionRectMask : QtGui.QAction
    actionMask : QtGui.QAction

    Methods
    -------
    connecttriggers
        Connect toolbar action signals to give slots
    """

    def __init__(self):
        super(Toolbar, self).__init__()

        self.actionRun_SlicePreview = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_50.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionRun_SlicePreview.setIcon(icon)
        self.actionRun_SlicePreview.setToolTip('Slice preview')

        self.actionRun_MultiSlicePreview = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_62.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionRun_MultiSlicePreview.setIcon(icon)
        self.actionRun_MultiSlicePreview.setToolTip('Multi-slice preview')

        self.actionRun_3DPreview = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_42.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionRun_3DPreview.setIcon(icon)
        self.actionRun_3DPreview.setToolTip('3D preview')

        self.actionRun_FullRecon = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_34.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionRun_FullRecon.setIcon(icon)
        self.actionRun_FullRecon.setToolTip('Full reconstruction')

        self.actionMBIR = QtGui.QWidgetAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_06.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionMBIR.setIcon(icon)
        self.actionMBIR.setToolTip('MBIR reconstruction')
        self.toolbuttonMBIR = QtGui.QToolButton(parent=self)
        self.toolbuttonMBIR.setPopupMode(QtGui.QToolButton.InstantPopup)
        self.actionMBIR.setDefaultWidget(self.toolbuttonMBIR)
        self.actionMBIR.setCheckable(True)
        self.toolbuttonMBIR.setDefaultAction(self.actionMBIR)

        self.actionCenter = QtGui.QWidgetAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_28.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionCenter.setIcon(icon)
        self.actionCenter.setToolTip('Center of rotation detection')
        self.toolbuttonCenter = QtGui.QToolButton(parent=self)
        self.toolbuttonCenter.setPopupMode(QtGui.QToolButton.InstantPopup)
        self.actionCenter.setDefaultWidget(self.toolbuttonCenter)
        self.actionCenter.setCheckable(True)
        self.toolbuttonCenter.setDefaultAction(self.actionCenter)

        self.actionPolyMask = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_05.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionPolyMask.setIcon(icon)
        self.actionPolyMask.setText("Polygon mask")

        self.actionCircMask = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_05.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionCircMask.setIcon(icon)
        self.actionCircMask.setText("Circular mask")

        self.actionRectMask = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_05.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionRectMask.setIcon(icon)
        self.actionRectMask.setText('Rectangular mask')

        self.actionMask = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_03.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.actionMask.setIcon(icon)

        maskmenu = QtGui.QMenu(self)
        maskmenu.addAction(self.actionRectMask)
        maskmenu.addAction(self.actionCircMask)
        maskmenu.addAction(self.actionPolyMask)
        toolbuttonMasking = QtGui.QToolButton(self)
        toolbuttonMasking.setDefaultAction(self.actionMask)
        toolbuttonMasking.setMenu(maskmenu)
        toolbuttonMasking.setPopupMode(QtGui.QToolButton.InstantPopup)
        toolbuttonMaskingAction = QtGui.QWidgetAction(self)
        toolbuttonMaskingAction.setDefaultWidget(toolbuttonMasking)

        self.actionROI = QtGui.QAction(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_60.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionROI.setIcon(icon)
        self.actionROI.setToolTip('Select region of interest')


        self.openButton = QtGui.QToolButton(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_55.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.openButton.setIcon(icon)
        self.openButton.setToolTip('Open flats/darks')
        openMenu = QtGui.QMenu()
        self.openFlats = QtGui.QAction('Open flats', openMenu)
        self.openDarks = QtGui.QAction('Open darks', openMenu)
        openMenu.addActions([self.openFlats, self.openDarks])

        self.openButton.setMenu(openMenu)
        self.openButton.setPopupMode(QtGui.QToolButton.ToolButtonPopupMode.InstantPopup)
        self.openButton.setArrowType(QtCore.Qt.NoArrow)

        self.setIconSize(QtCore.QSize(32, 32))

        self.addAction(self.actionRun_FullRecon)
        self.addAction(self.actionRun_SlicePreview)
        self.addAction(self.actionRun_MultiSlicePreview)
        self.addAction(self.actionRun_3DPreview)
        self.addAction(self.actionCenter)
        self.addAction(self.actionROI)
        self.addWidget(self.openButton)

        # mbir functionality should be replaced by remote workflow (dask)
        # self.addAction(self.actionMBIR)

        # self.addAction(toolbuttonMaskingAction)


    def connectTriggers(self, slicepreview, multislicepreview, preview3D, fullrecon, center, roiselection, mbir,
                        openflats, opendarks):

        """
        Connect toolbar action signals to give slots

        Parameters
        ----------
        slicepreview : QtCore.Slot
            Slot to connect actionRun_SlicePreview
        preview3D : QtCore.Slot
            Slot to connect actionRun_3DPreview
        fullrecon : QtCore.Slot
            Slot to connect actionRun_FullRecon
        center : QtCore.Slot
            Slot to connect actionCenter
        """

        self.actionRun_SlicePreview.triggered.connect(slicepreview)
        self.actionRun_MultiSlicePreview.triggered.connect(multislicepreview)
        self.actionRun_3DPreview.triggered.connect(preview3D)
        self.actionRun_FullRecon.triggered.connect(fullrecon)
        self.actionCenter.toggled.connect(center)
        self.actionMBIR.toggled.connect(mbir)
        self.actionROI.triggered.connect(roiselection)
        self.openFlats.triggered.connect(openflats)
        self.openDarks.triggered.connect(opendarks)

class ReconManager(QtGui.QSplitter):
    """
    Widget to show reconstruction queue on leftwidget tab. Also has actual queue used to hold recon jobs

    Attributes
    ----------
    queue_form : QtTui.QStackedWidget
        widget to show dictionary representation of pipeline for each reconstruction in queue
    queue : featurewidgets.FeatureManager
        widget to hold featurewidgets representing reconstruction jobs
    manager : featurewidgets.FeatureManager
        Widget for holding recon jobs and keeping their order

    """


    def __init__(self, *args, **kwargs):

        super(ReconManager, self).__init__(*args, **kwargs)
        queue_ui = QUiLoader().load('xicam/gui/tomographyqueue.ui')
        self.queue_form = QtGui.QStackedWidget()
        queue_ui.functionsList.setAlignment(QtCore.Qt.AlignBottom)
        queue_ui.moveDownButton.setToolTip('Move selected job down in queue')
        queue_ui.moveUpButton.setToolTip('Move selected job up in queue')
        self.manager = fw.FeatureManager(queue_ui.functionsList, self.queue_form, blank_form='Click items below to see reconstruction jobs on queue.')
        queue_ui.moveDownButton.clicked.connect(self.moveDown)
        queue_ui.moveUpButton.clicked.connect(self.moveUp)

        self.setOrientation(QtCore.Qt.Vertical)
        self.addWidget(self.queue_form)
        self.addWidget(queue_ui)

        # queue for reconstructions
        self.recon_queue = deque()

    def swapQueue(self, idx1, idx2):
        """
        Function to swap jobs sitting in queue

        Parameters
        ----------
        idx1: int
            index of first job
        idx2: int
            index of second job
        """

        idx1 -= 1
        idx2 -= 1

        tmp = self.recon_queue[idx1]
        self.recon_queue[idx1] = self.recon_queue[idx2]
        self.recon_queue[idx2] = tmp
        del(tmp)

    def delQueueJob(self, idx):
        """
        Function to delete job on queue.

        Parameters
        ----------
        idx: int
            index of job to delete
        """

        idx -= 1

        job = self.recon_queue[idx]
        self.recon_queue.remove(job)


    def addRecon(self, args):
        """
        Adds reconstruction job to queue. Renders parameters and function pipeline in a data tree

        Parameters
        ----------
        args: 7-tuple
            Tuple with args for a reconstruction. They are:
            1). datawidget: Datawidget containing data
            2). tuple: function pipeline, and reconstruction params (like theta, COR)
            3). type: projections to reconstruct
            4). tuple: sinograms to reconstruct
            5). int: sinograms per reconstruction iteration
            6). tuple: width dimensions for reconstruction
            7). int: cpus available
        """

        name = os.path.basename(args[0].path)
        widget = fw.FeatureWidget(name, checkable=False)
        widget.previewButton.hide()
        widget.line.hide()
        widget.sigDelete.connect(self.reconDeleted)

        form = DataTreeWidget()
        data = args[4]
        form.setData(data, hideRoot=True)
        form.header().hide()
        form.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        form.setSelectionBehavior(QtGui.QAbstractItemView.SelectItems)

        widget.form = form

        self.queue_form.addWidget(form)
        self.manager.addFeature(widget)


    def removeRecon(self, idx):
        """
        Function to remove job from queue

        Parameters
        ----------
        idx: int
            index of job to be removed

        """

        feature = self.manager.features[idx]
        self.manager.removeFeature(feature)

    def reconDeleted(self, funcwidget):
        """
        Slot to connect to user deleting job on queue

        Parameters
        ----------
        funcwidget: featurewidgets.FunctionWidget
            functionwidget to be deleted
        """

        idx = self.manager.features.index(funcwidget)
        self.delQueueJob(idx)

    def moveUp(self):
        """
        Slot to connect to user swapping recon job and job above it
        """

        idx1 = self.manager.features.index(self.manager.selectedFeature)
        idx2 = self.manager.features.index(self.manager.nextFeature)

        if idx1 == 0 or idx2 == 0:
            return
        else:
            self.manager.swapFeatures(self.manager.selectedFeature, self.manager.nextFeature)
            self.swapQueue(idx1, idx2)

    def moveDown(self):
        """
        Slot to connect to user swapping recon job and job below it
        """
        idx1 = self.manager.features.index(self.manager.selectedFeature)
        idx2 = self.manager.features.index(self.manager.previousFeature)

        if idx1 == 0 or idx2 == 0:
            return
        else:
            self.manager.swapFeatures(self.manager.selectedFeature, self.manager.previousFeature)
            self.swapQueue(idx1, idx2)



import base


def runtest():
    import numpy as np

    img = np.random.random((100, 100, 100))
    EZTest.setImage(img)

    hist = np.histogram(img, 100)
    EZTest.plot(hist[1][:-1], hist[0])


def opentest(filepaths):
    import fabio
    for filepath in filepaths:
        img = fabio.open(filepath).data
        EZTest.setImage(img)


EZTest = base.EZplugin(name='EZTest',
                       toolbuttons=[('xicam/gui/icons_34.png', runtest)],
                       parameters=[{'name': 'Test', 'value': 10, 'type': 'int'},
                                   {'name': 'Fooo', 'value': True, 'type': 'bool'}],
                       openfileshandler=opentest)






#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import warnings
import threading

import numpy
import scipy.optimize
import pyFAI


FIT_PARAMETER = ['wavelength',
                 'distance',
                 'center_x',
                 'center_y',
                 'tilt',
                 'rotation']
LN_2 = numpy.log(2)

x = []
y = []


def gauss(x_max, y_max, fwhm, x_array):
    return y_max * numpy.exp(-LN_2 * ((x_array - x_max) / fwhm * 2) ** 2)


def lorentz(x_max, y_max, fwhm, x_array):
    return y_max * 1. / (1 + ((x_array - x_max) / fwhm * 2) ** 2)


def pseudo_voigt(x_max, y_max, fwhm, eta, x_array):
    return eta * gauss(x_max, y_max, fwhm, x_array) + \
           (1 - eta) * lorentz(x_max, y_max, fwhm, x_array)


def fit_maxima(x_data, y_data):
    # print(x_data,y_data)
    # error function
    err = lambda p, x, y: y - (p[0] + p[1] * x + pseudo_voigt(p[2], p[3], p[4], p[5], x))

    # approximate the fitting parameter
    y_min = y_data.min()
    y_max = y_data.max() - y_min
    x_max = x_data[y_data.argmax()]
    parameter = [y_min, 1E-6, x_max, y_max, 1, 1]
    args = (x_data, y_data)

    # fit PSV with linear background to data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        parameter, info = scipy.optimize.leastsq(err, parameter, args)

    # return maximum position if valid solution, otherwise nan
    maxima = parameter[2]
    if info != 5 and maxima > x_data.min() and maxima < x_data.max():
        return parameter[2]
    else:
        return numpy.nan


def _error_function(parameter, arguments):
    '''
    The error functions returns the distances using the current parameter.
    This is a helper function for the fitting process.
    '''
    geometry, d_spacings, maxima, selected_parameter = arguments
    mask = [sel in selected_parameter for sel in FIT_PARAMETER]
    param = numpy.array(get_fit2d(geometry))
    print 'update:',param
    param[numpy.array(mask)] = parameter[numpy.array(mask)]
    set_fit2d(geometry,*param)
    return peak_distance(geometry, d_spacings, maxima)

def get_fit2d(geometry):
    gdict = geometry.getFit2D()
    return [geometry.get_wavelength()*1e-10,gdict['directDist'],gdict['centerX'],gdict['centerY'],gdict['tilt'],gdict['tiltPlanRotation']]

def set_fit2d(geometry, wavelength, distance, center_x, center_y, tilt, rotation):
    geometry.set_wavelength(wavelength / 1e-10)
    geometry.setFit2D(distance, center_x, center_y, tilt, rotation)
    return geometry


def fit_geometry(geometry, maxima, d_spacings, selected_parameter):
    args = [geometry, d_spacings, maxima[::-1], selected_parameter]
    mask = numpy.array([sel in selected_parameter for sel in FIT_PARAMETER])
    start_parameter = numpy.array(get_fit2d(geometry))

    if len(start_parameter) >= len(maxima[0]):
        raise Exception('More variables then fit points.')

    _, cov_x, info, _, _ = scipy.optimize.leastsq(_error_function,
                                                  start_parameter,
                                                  args,
                                                  full_output=True)
    fit_parameter = start_parameter

    dof = len(maxima[0]) - len(start_parameter)
    chisq = (info['fvec'] ** 2).sum()
    info_out = {'calls': info['nfev'],
                'dof': dof,
                'sum_chi_square': chisq}
    info_out['variance_residuals'] = chisq / float(dof)

    for i, name in enumerate(selected_parameter):
        if cov_x is not None:
            ase = numpy.sqrt(cov_x[i, i]) * numpy.sqrt(chisq / float(dof))
        else:
            ase = numpy.NAN
        info_out['ase_%s' % name] = ase
        percent_ase = round(ase / (abs(fit_parameter[i]) * 0.01), 3)
        info_out['percent_ase_%s' % name] = percent_ase
        info_out[name] = fit_parameter[i]

    return info_out


def radial_array(beam_center, shape):
    '''
    Returns array with shape where every point is the radial distance to the
    beam center in pixel.
    '''
    pos_x, pos_y = numpy.meshgrid(numpy.arange(shape[1]),
                                  numpy.arange(shape[0]))
    return numpy.sqrt((pos_x - beam_center[0]) ** 2 + (pos_y - beam_center[1]) ** 2)


def ring_maxima(geometry, d_spacing, image, radial_pos, step_size):
    '''
    Returns x and y arrays of maximum positions [pixel] found on the ring
    defined by pyFAI geometry and d_spacing on the given image
    '''

    # calculate circle positions along the ring with step_size
    tth = 2 * numpy.arcsin(geometry.get_wavelength() / (2e-10 * d_spacing))
    radius = (geometry.get_dist() * numpy.tan(tth)) / geometry.get_pixel1()
    center = (geometry.getFit2D()['centerX'], geometry.getFit2D()['centerY'])
    alpha = numpy.arange(0, numpy.pi * 2, step_size / float(radius))
    circle_x = numpy.round(center[1] + numpy.sin(alpha) * radius)
    circle_y = numpy.round(center[0] + numpy.cos(alpha) * radius)

    # calculate roi coordinates    
    half_step = int(numpy.ceil(step_size / 2.))
    x_0 = circle_x - half_step
    x_1 = circle_x + half_step
    y_0 = circle_y - half_step
    y_1 = circle_y + half_step

    # mask out rois which are not complete  inside the image border
    mask = numpy.where((x_0 >= 0) & (y_0 >= 0) &
                       (y_1 < image.shape[0]) & (x_1 < image.shape[1]))
    x_0 = x_0[mask]
    x_1 = x_1[mask]
    y_0 = y_0[mask]
    y_1 = y_1[mask]

    maxima_x = []
    maxima_y = []
    for i in range(len(x_0)):
        roi = image[int(y_0[i]):int(y_1[i]), int(x_0[i]):int(x_1[i])]
        pos = radial_pos[int(x_0[i]):int(x_1[i]), int(y_0[i]):int(y_1[i])]
        if roi.size < half_step ** 2: continue

        # calculate roi histogram
        try:
            x_hist, y_hist, _, _ = pyFAI.ext.histogram.histogram(pos, roi, step_size)
        except AssertionError: continue

        # fit the radial maximum of the histogram
        maximum = fit_maxima(x_hist, y_hist)
        if maximum is numpy.nan: continue

        # DEBUG
        # rect = pylab.Rectangle((x_0[i], y_0[i]),
        # abs(x_0[i]-x_1[i]),
        #                       abs(y_0[i]-y_1[i]), fc='none')
        #pylab.gca().add_patch(rect)

        # calculate the pixel position of the maximum
        scale = float(maximum / radius)
        maxima_x.append(center[1] + (x_0[i] + half_step - center[1]) * scale)
        maxima_y.append(center[0] + (y_0[i] + half_step - center[0]) * scale)

        # DEBUG
        #pylab.plot(center[0] + (x_0[i] + half_step - center[0]) * scale, 
        #           center[1] + (y_0[i] + half_step - center[1]) * scale, '*')

    return numpy.array(maxima_x), numpy.array(maxima_y), radial_pos


def peak_distance(geometry, d_spacings, (x_peaks, y_peaks)):
    """
    Returns the minimum distances in 2*theta between given peaks in pixel and a 
    d_spacings in angstrom using the pyFAI geometry.
    """
    wavelength = geometry.get_wavelength()
    tth_rings = 2 * numpy.arcsin(wavelength / (2.0e-10 * d_spacings))
    distance = [numpy.abs(tth_rings - tth).min() \
                for tth in geometry.tth(y_peaks, x_peaks)]
    return numpy.array(distance)

def circle_center_distance(center, x_array, y_array):
    """
    Returns the algebraic distances between the 2D points in x_array, y_array
    and the mean circle center.
    """
    r_array = numpy.sqrt((x_array - center[0]) ** 2 + (y_array - center[1]) ** 2)
    return r_array - r_array.mean()


def fit_circle(x_array, y_array):
    """
    Returns the center (x, y) and the radius of the fitted circle defined by
    the points in x_array and y_array.
    """
    # coordinates of the barycenter
    xc_m = numpy.mean(x_array)
    yc_m = numpy.mean(y_array)

    # fit circle center
    center, _ = scipy.optimize.leastsq(circle_center_distance,
                                       (xc_m, yc_m),
                                       (x_array, y_array))

    # calculate radius
    radius = numpy.sqrt((x_array - center[0]) ** 2 +
                        (y_array - center[1]) ** 2).mean()
    return center, radius


def quick_calibration(wavelength, pixel_size, d_spacing, x_array, y_array):
    '''
    Returns pyFAI geometry by fitting a circle to a d-spacing ring.  
    '''
    center, radius = fit_circle(x_array, y_array)
    tth = 2 * numpy.arcsin(wavelength / (2.0e-10 * d_spacing))
    sdd = pixel_size[0] * radius * numpy.cos(tth) * 1. / numpy.sin(tth)
    return PyFAIGeometry(sdd,
                         center[1] * pixel_size[1],
                         center[0] * pixel_size[0],
                         1e-6,
                         1e-6,
                         1e-6,
                         pixel_size[0],
                         pixel_size[1],
                         wavelength=wavelength)


# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
#   added at LBL    # 
def set_PyFAIGeometry(agb):
    wave = agb.wavelength_ * 1.0E-10
    a = PyFAIGeometry(
        agb.sdd_,
        agb.center_[1] * agb.pixel_size_[1],
        agb.center_[0] * agb.pixel_size_[0],
        1e-6,
        1e-6,
        1e-6,
        agb.pixel_size_[0],
        agb.pixel_size_[1],
        wavelength=wave)
    return a


#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class PyFAIGeometry(pyFAI.geometry.Geometry):
    def set_fit2d(self,
                  wavelength,
                  distance,
                  center_x,
                  center_y,
                  tilt,
                  rotation):
        self.set_wavelength(wavelength * 1e-10)
        self.setFit2D(distance, center_x, center_y, tilt, rotation)

    def get_fit2d(self):
        param_dict = self.getFit2D()
        return [self.get_wavelength() * 1e10,
                param_dict['directDist'],
                param_dict['centerX'],
                param_dict['centerY'],
                param_dict['tilt'],
                param_dict['tiltPlanRotation']]


################################################################################

class FitThread(threading.Thread):
    def __init__(self, geometry, d_spacings, image, selected_parameter, step_size):
        threading.Thread.__init__(self)
        self.geometry = geometry
        self.d_spacings = d_spacings
        self.image = image
        self.selected_paramter = selected_parameter
        self.step_size = step_size
        self.status = [0, len(d_spacings)]
        self.info = {}
        self.circle_patches = []

    def run(self):
        # calculate maxima for every d_spacing 
        center = (self.geometry.getFit2D()['centerX'],
                  self.geometry.getFit2D()['centerY'])
        radial_pos = radial_array(center, self.image.shape)
        x_data, y_data, = [], []

        # calculate maxima for every d_spacing 
        for d_spacing in self.d_spacings:
            maxima_x, maxima_y, radial_pos = ring_maxima(self.geometry,
                                                         d_spacing,
                                                         self.image,
                                                         radial_pos,
                                                         self.step_size)
            x_data.extend(maxima_x)
            y_data.extend(maxima_y)

            # create circle patch for every found maxima
            # for i in range(len(maxima_x)):
            # c_p = pylab.Circle((maxima_x[i], maxima_y[i]), 10, ec='red', fc='none')
            # self.circle_patches.append(c_p)

            self.status[0] += 1

        # start fit
        try:
            info = fit_geometry(self.geometry,
                                (numpy.array(x_data), numpy.array(y_data)),
                                self.d_spacings,
                                self.selected_paramter)
            info.update({'error': False, 'error_msg': ''})
            self.info = info
        except Exception as e:
            self.info = {'error': True, 'error_msg': e.message}

    def get_status(self):
        return tuple(self.status)

    def get_info(self):
        return self.info

    def get_circle_patches(self):
        return self.circle_patches


#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
#   added at LBL    # 

def dpdak_saxs_calibration(img, geometry, d_spacings, fit_param, step_size):
    fit_thread = FitThread(geometry, d_spacings, img, fit_param, step_size)
    fit_thread.start()
    params = geometry.get_fit2d()
    return params[1] / 1000., [params[2], params[3]]


def saxs_calibration(img, agb):
    max_iter = 10
    tol = 10E-04

    # d-spacing for Silver Behenate
    d_spacings = numpy.array([58.367, 29.1835, 19.45567, 14.59175, \
                              11.6734, 9.72783, 8.33814, 7.29587])

    # step size (same as in dpdak GUI)
    step_size = 49

    # Set PyFAI Geometry
    geometry = set_PyFAIGeometry(agb)

    # selcte parameters to fit
    fit_param = ['distance', 'center_x', 'center_y', 'tilt']

    # Run calibration
    for i in range(10):
        fit_thread = FitThread(geometry, d_spacings, img, fit_param, step_size)
        fit_thread.start()
        fit_thread.join()

    # plt.plot(x, y)
    # plt.show()

    # Update calibrated data
    params = geometry.get_fit2d()
    agb.setSDD(params[1] / 1000.)
    agb.setCenter(params[2:4])
    return agb


#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


#if __name__ == "__main__":
# test_all()



import numpy as np
import itertools


def reciprocalvectors(a, b, c, order=2):
    V = abs(np.dot(a, np.cross(b, c)))
    mra = 2 * np.pi / V * np.cross(b, c)
    mrb = 2 * np.pi / V * np.cross(c, a)
    mrc = 2 * np.pi / V * np.cross(a, b)
    mi = np.vstack([mra, mrb, mrc])
    combs = itertools.product(range(-order, order + 1), repeat=3)
    vecs = [np.sum((mi.T * np.array(comb)).T, axis=0) for comb in combs]
    return vecs

def combs_generator(x,y,z):
    for i in range(-x/2,x/2+1):
        for j in range(-y/2,y/2+1):
            for k in range(0,z+1):
                yield (i,j,k)

def latticevectors(a, b, c, zoffset, maxreps=100, repetitions=None,scaling=1.):
    if repetitions is None: repetitions = [0,0,0]
    max_rep = [5,5,5]
    repet_list = list(repetitions)
    for n in range(0,2):
        repet_list[n] = min(repet_list[n],max_rep[n])
    repetitions = tuple(repet_list)
    mi = np.vstack([a, b, c])
    combs = combs_generator(*repetitions) #itertools.product(range(-order, order + 1), repeat=3)
    # vecs = [np.sum((mi.T * np.array(comb)).T, axis=0) for comb in combs]
    # vecs = [zoffsetvec(vec,zoffset) for vec in vecs if vec[2] >= 0 and np.sqrt(vec[0]**2+vec[1]**2)<=maxr and vec[2]<=maxz]
    vecs = [zoffsetvec(np.sum((mi.T * np.array(comb)*scaling).T, axis=0),zoffset) for comb in combs]
    vecs = sorted(vecs,key=lambda v: np.vdot(v,v))[:maxreps]
    return vecs

def zoffsetvec(vec,zoffset=0):
    vec[2]=vec[2]+zoffset
    return vec


def latticelines(a, b, c, zoffset, maxreps=100, repetitions=None, scaling=1.):
    vecs = latticevectors(a, b, c, zoffset, maxreps, repetitions, scaling)

    lines = []
    for vec in vecs:
        lines.extend([(vec, vec + a), (vec, vec - a), (vec, vec + b), (vec, vec - b), (vec, vec + c), (vec, vec - c)])

    return lines


def vecs2angles(v):
    v = np.array(v)
    vproj = v.copy()
    vproj[:, 2] = 0
    return np.arctan2(vproj[:, 1], vproj[:, 0]) / np.pi * 180.


if __name__ == '__main__':
    a = [1, 0, 0]
    b = [.5, np.sqrt(3) / 2, 0]
    c = [0, 0, 1]
    recipvecs = np.array(reciprocalvectors(a, b, c))

    print(list(set(vecs2angles(recipvecs))))

    exit(0)



import os
import sys
import inspect
import h5py
import tifffile
import glob
import numpy as np
from fabio.fabioimage import fabioimage
import fabio
import pyFAI
from pyFAI import detectors
import logging
from . import msg
import pyfits
# from nexusformat import nexus as nx
from collections import OrderedDict
import re

if fabio._version.MAJOR==0 and fabio._version.MINOR<5:
    from PySide import QtGui
    msgBox = QtGui.QMessageBox()
    msgBox.setText("The FabIO package is an older version and must be updated. Xi-cam can try to install this for you.")
    msgBox.setInformativeText("Would you like to install FabIO?")
    msgBox.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
    msgBox.setDefaultButton(QtGui.QMessageBox.Yes)

    response = msgBox.exec_()
    if response == QtGui.QMessageBox.Yes:
        import pip
        failure = pip.main(['install','--upgrade','fabio'])

        from xicam.dialogs import infodialog
        if failure:
            infodialog("The FabIO package could not be updated. Please update FabIO manually (i.e. 'pip install --upgrade fabio').",'Update Failed')
        else:
            infodialog("The FabIO package was updated. Please restart Xi-cam.",'Update Success!')
            exit(0)

from fabio import fabioutils, edfimage, tifimage, fabioformats

h5classes = list()
tiffclasses = list()

def register_tiffclass(cls):
    global tiffclasses
    tiffclasses.append(cls)
    return cls

def register_h5class(cls):
    global h5classes
    h5classes.append(cls)
    return cls


class xicamtiffimage(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['tiff', 'tif']

    def read(self, filename, frame=None):
        for tiff in tiffclasses:
            if hasattr(tiff,'validate'): # check which class preferably based on the validate staticmethod
                try:
                    tiff.validate(filename, frame)
                except Exception as ex:
                    continue
            try: # if there isn't one, try to read with this class
                return xicamtiffimage._instantiate_read(tiff,filename,frame)
            except Exception as ex:
                continue

        # if custom classes fail, use built-in class
        return xicamtiffimage._instantiate_read(fabio.tifimage.tifimage, filename, frame)

    @staticmethod
    def _instantiate_read(cls,filename,frame):
        fabh5 = cls()
        fabh5.filename = filename
        return fabh5.read(filename, frame)

fabio.openimage.MAGIC_NUMBERS.insert(0,(b"\x49\x49", 'xicamtiff'))


class hdf5image(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['hdf', 'h5', 'hdf5']

    # A proxy class which defers to specific hdf5 schema classes
    def read(self, filename, frame=None):
        for h5cls in h5classes:
            if hasattr(h5cls,'validate'): # check which class preferably based on the validate staticmethod
                try:
                    h5cls.validate(filename, frame)
                except Exception as ex:
                    continue
            try: # if there isn't one, try to read with this class
                return hdf5image._instantiate_read(h5cls,filename,frame)
            except Exception as ex:
                continue

    @staticmethod
    def _instantiate_read(cls,filename,frame):
        fabh5 = cls()
        fabh5.filename = filename
        return fabh5.read(filename, frame)

fabio.openimage.MAGIC_NUMBERS.insert(0,(b"\x89\x48\x44\x46", 'hdf5'))


@register_h5class
class ALS832H5image(fabioimage):
    """
    Fabio Image class for ALS Beamline 8.3.2 HDF5 Datasets
    """
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['h5']

    def __init__(self, data=None, header=None, transpose=False):
        super(ALS832H5image, self).__init__(data=data, header=header)
        self.frames = None
        self.currentframe = 0
        self.header = None
        self._h5 = None
        self._dgroup = None
        self._flats = None
        self._darks = None

    # Context manager for "with" statement compatibility
    def __enter__(self, *arg, **kwarg):
        return self

    def __exit__(self, *arg, **kwarg):
        self.close()

    def _readheader(self, f):
        if self._h5 is not None:
            self.header = dict(self._h5.attrs)
            self.header.update(**self._dgroup.attrs)

    @staticmethod
    def validate(f, frame=None):
        h5 = h5py.File(f, 'r')
        header = dict(h5.attrs)
        assert header['facility'] == 'als'
        assert header['end_station'] == 'bl832'

    def read(self, f, frame=None):
        self.filename = f
        if frame is None:
            frame = 0
        if self._h5 is None:

            # Check header for unique attributes
            try:
                self._h5 = h5py.File(self.filename, 'r+')
                self._dgroup = self._finddatagroup(self._h5)
                self.readheader(f)
                if self.header['facility'] != 'als' or self.header['end_station'] != 'bl832':
                    raise H5ReadError
            except KeyError:
                raise H5ReadError

            self.frames = [key for key in self._dgroup.keys() if 'bak' not in key and 'drk' not in key]
        dfrm = self._dgroup[self.frames[frame]]
        self.currentframe = frame
        self.data = dfrm[0]
        return self

    def change_dataset_attribute(self, key, value):
        self._dgroup.attrs.modify(key, value)

    def _finddatagroup(self, h5object):
        keys = h5object.keys()
        if len(keys) == 1:
            if isinstance(h5object[keys[0]], h5py.Group):
                group_keys = h5object[keys[0]].keys()
                if isinstance(h5object[keys[0]][group_keys[0]], h5py.Dataset):
                    return h5object[keys[0]]
                else:
                    return self._finddatagroup(h5object[keys[0]])
            else:
                raise H5ReadError('Unable to find dataset group')
        else:
            raise H5ReadError('Unable to find dataset group')

    @property
    def flats(self):
        if self._flats is None:
            self._flats = OrderedDict()
            for key in sorted(self._dgroup.keys()):
                if 'bak' in key:
                    self._flats[key] = self._dgroup[key][0]
        return self._flats

    @property
    def darks(self):
        if self._darks is None:
            if self._darks is None:
                self._darks = OrderedDict()
                for key in sorted(self._dgroup.keys()):
                    if 'drk' in key:
                        self._darks[key] = self._dgroup[key][0]
            return self._darks
        return self._darks

    def flatindices(self):
        i0 = int(self.header['i0cycle'])
        nproj = len(self)
        if i0 > 0:
            indices = list(range(0, nproj, i0))
            if indices[-1] != nproj - 1:
                indices.append(nproj - 1)
        elif i0 == 0:
            indices = [0, nproj - 1]
        return indices

    @property
    def nframes(self):
        return len(self.frames)

    @nframes.setter
    def nframes(self, n):
        pass

    # def getsinogram(self, idx=None):
    #     if idx is None: idx = self.data.shape[0]//2
    #     self.sinogram = np.vstack([frame for frame in map(lambda x: self._dgroup[self.frames[x]][0, idx],
    #                                                               range(self.nframes))])
    #     return self.sinogram

    def __getitem__(self, item):
        s = []
        if not isinstance(item, tuple) and not isinstance(item, list):
            item = (item,)
        for n in range(3):
            if n == 0:
                stop = len(self)
            elif n == 1:
                stop = self.data.shape[0]
            elif n == 2:
                stop = self.data.shape[1]
            if n < len(item) and isinstance(item[n], slice):
                start = item[n].start if item[n].start is not None else 0
                step = item[n].step if item[n].step is not None else 1
                stop = item[n].stop if item[n].stop is not None else stop
            elif n < len(item) and isinstance(item[n], int):
                if item[n] < 0:
                    start, stop, step = stop + item[n], stop + item[n] + 1, 1
                else:
                    start, stop, step = item[n], item[n] + 1, 1
            else:
                start, step = 0, 1

            s.append((start, stop, step))

        for n, i in enumerate(range(s[0][0], s[0][1], s[0][2])):
            _arr = self._dgroup[self.frames[i]][0, slice(*s[1]), slice(*s[2])]
            if n == 0:  # allocate array
                arr = np.empty((len(range(s[0][0], s[0][1], s[0][2])), _arr.shape[0], _arr.shape[1]))
            arr[n] = _arr
        if arr.shape[0] == 1:
            arr = arr[0]
        return np.squeeze(arr)

    def __len__(self):
        return self.nframes

    def getframe(self, frame=0):
        self.data = self._dgroup[self.frames[frame]][0]
        return self.data


    def next(self):
        if self.currentframe < self.__len__() - 1:
            self.currentframe += 1
        else:
            raise StopIteration
        return self.getframe(self.currentframe)

    def previous(self):
        if self.currentframe > 0:
            self.currentframe -= 1
            return self.getframe(self.currentframe)
        else:
            raise StopIteration

    def close(self):
        self._h5.close()


class nexusimage(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['hdf']

    def read(self, f, frame=None):
        # nxroot = nx.nxload(f)
        # # print nxroot.tree
        # if hasattr(nxroot, 'entry'):
        #     if hasattr(nxroot.entry, 'data'):
        #         if hasattr(nxroot.entry.data,'data'):
        #             self._dgroup = nxroot.entry.data.data

        self.filename = f
        if frame is None:
            frame = 0
        if self._h5 is None:
            # Check header for unique attributes
            self._h5 = h5py.File(self.filename, 'r+')
            self._dgroup = self._h5['entry']['data']['data']
            self.readheader(f)

            self.frames = range(self._dgroup.shape[0])

        dfrm = self._dgroup[self.frames[frame]]
        self.currentframe = frame
        self.data = dfrm



        return self

    @staticmethod
    def validate(f, frame=None):
        h5 = h5py.File(f, 'r')
        assert list(h5.keys())[0] == 'entry'
        assert 'data' in list(h5['entry'].keys())
        assert list(h5['entry']['data'])[0] == 'data'

    def __init__(self, data=None , header=None):
        super(nexusimage, self).__init__(data=data, header=header)
        self.frames = None
        self.currentframe = 0
        self.header = None
        self._h5 = None
        self._dgroup = None
        self._flats = None
        self._darks = None

    # Context manager for "with" statement compatibility
    def __enter__(self, *arg, **kwarg):
        return self

    def change_dataset_attribute(self, key, value):
        self._dgroup.attrs.modify(key, value)

    def __exit__(self, *arg, **kwarg):
        self.close()

    def _readheader(self, f):
        #not really useful at this point
        if self._h5 is not None:
            self.header=dict(self._h5.attrs)
            self.header.update(**self._dgroup.attrs)

    @property
    def flats(self):
        return self._flats

    @property
    def darks(self):
        return self._darks

    def flatindices(self):
        nproj = len(self)
        return [0, nproj - 1]

    @property
    def nframes(self):
        return len(self.frames)

    @nframes.setter
    def nframes(self, n):
        pass

    # def getsinogram(self, idx=None):
    #     if idx is None: idx = self.data.shape[0]//2
    #     self.sinogram = np.vstack([frame for frame in map(lambda x: self._dgroup[self.frames[x]][0, idx],
    #                                                               range(self.nframes))])
    #     return self.sinogram

    def __getitem__(self, item):
        s = []
        if not isinstance(item, tuple) and not isinstance(item, list):
            item = (item, )
        for n in range(3):
            if n == 0:
                stop = len(self)
            elif n == 1:
                stop = self.data.shape[0]
            elif n == 2:
                stop = self.data.shape[1]
            if n < len(item) and isinstance(item[n], slice):
                    start = item[n].start if item[n].start is not None else 0
                    step = item[n].step if item[n].step is not None else 1
                    stop = item[n].stop if item[n].stop is not None else stop
            elif n < len(item) and isinstance(item[n], int):
                    if item[n] < 0:
                        start, stop, step = stop + item[n], stop + item[n] + 1, 1
                    else:
                        start, stop, step = item[n], item[n] + 1, 1
            else:
                start, step = 0, 1

            s.append((start, stop, step))

        for n, i in enumerate(range(s[0][0], s[0][1], s[0][2])):
            _arr = self._dgroup[self.frames[i]][slice(*s[1]), slice(*s[2])]
            if n == 0:  # allocate array
                arr = np.empty((len(range(s[0][0], s[0][1], s[0][2])), _arr.shape[0], _arr.shape[1]))
            arr[n] = _arr
        if arr.shape[0] == 1:
            arr = arr[0]
        return np.squeeze(arr)

    def __len__(self):
        return self.nframes

    def getframe(self, frame=0):
        self.data = self._dgroup[self.frames[frame]]
        return self.data

    def next(self):
        if self.currentframe < self.__len__() - 1:
            self.currentframe += 1
        else:
            raise StopIteration
        return self.getframe(self.currentframe)

    def previous(self):
        if self.currentframe > 0:
            self.currentframe -= 1
            return self.getframe(self.currentframe)
        else:
            raise StopIteration

    def close(self):
        self._h5.close()


@register_tiffclass
class tomotifimage(fabioimage):

    """
    Fabio class for tiff images (specifically for tomography)
    """

    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['tif', 'tiff']

    def __init__(self, data=None, header=None):
        super(tomotifimage, self).__init__(data=data, header=header)
        self.frames = None
        self.currentframe = 0
        self._dgroup = None
        self.header = None
        self.flats = None
        self.darks = None

    @staticmethod
    def validate(f, frame=None):
        tiff = tifffile.imread(f)
        assert len(tiff.shape) > 2

    def read(self, f, frame=None):
        self._dgroup = tifffile.imread(f)
        self.data = self._dgroup[0]
        self.frames = range(self._dgroup.shape[0])
        return self

    def getframe(self, frame=0):
        self.data = self._dgroup[frame]
        return self.data

    def __getitem__(self, item):
        return self._dgroup[item]

    def __len__(self):
        return self._dgroup.shape[0]

    def flatindices(self):
        nproj = len(self)
        return [0, nproj - 1]

    def next(self):
        if self.currentframe < self.__len__() - 1:
            self.currentframe += 1
        else:
            raise StopIteration
        return self.getframe(self.currentframe)

    def previous(self):
        if self.currentframe > 0:
            self.currentframe -= 1
            return self.getframe(self.currentframe)
        else:
            raise StopIteration

    def close(self):
        pass

class npyimage(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['npy']

    def read(self, f, frame=None):
        self.data = np.load(f)
        return self


class hipgisaxsimage(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['out']

    def read(self, f, frame=None):
        data = np.loadtxt(f)
        data = (data / data.max() * ((2 ** 32) - 1)).astype(np.uint32).copy()
        self.data = data
        return self


class fitsimage(fabioimage):
    DESCRIPTION = "FITS file format from astronomy"

    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ["fits"]

    def read(self, f, frame=None):
        self.data = np.rot90(np.fliplr(pyfits.open(f)[2].data), 2)
        return self


class gbimage(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['gb']

    def read(self, f, frame=None):
        data = np.fromfile(f, np.float32)
        if len(data) == 1475 * 1679:
            data.shape = (1679, 1475)
        elif len(data) == 981 * 1043:
            data.shape = (1043, 981)
        elif len(data) == 1475 * 195:
            data.shape = (195, 1475)
        self.data = data
        return self


class rawimage(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['raw']

    def read(self, f, frame=None):
        with open(f, 'r') as f:
            data = np.fromfile(f, dtype=np.int32)
        for name, detector in detectors.ALL_DETECTORS.iteritems():
            if hasattr(detector, 'MAX_SHAPE'):
                # print name, detector.MAX_SHAPE, imgdata.shape[::-1]
                if np.prod(detector.MAX_SHAPE) == len(data):  #
                    detector = detector()
                    msg.logMessage('Detector found: ' + name, msg.INFO)
                    break
            if hasattr(detector, 'BINNED_PIXEL_SIZE'):
                # print detector.BINNED_PIXEL_SIZE.keys()
                if len(data) in [np.prod(np.array(detector.MAX_SHAPE) / b) for b in
                                 detector.BINNED_PIXEL_SIZE.keys()]:
                    detector = detector()
                    msg.logMessage('Detector found with binning: ' + name, msg.INFO)
                    break
        data.shape = detector.MAX_SHAPE
        self.data = data
        return self


# @register_fabioclass
# class H5image(fabioimage):
#     """
#     HDF5 Fabio Image class (hack?) to allow for different internal HDF5 structures.
#     To create a fabimage for another HDF5 structure simply define the class in this module like any other fabimage
#     subclass and include 'H5image' somewhere in its name.
#     """
#
#     # This does not really work because fabio creates the instance of the class with not connection to the filename
#     # and only after instantiation does it call read. Therefore any bypasses at __new__ seem to be futile
#     # def __new__(cls, *args, **kwargs):
#     #     h5image_classes = [image for image in inspect.getmembers(sys.modules[__name__], inspect.isclass)
#     #                        if 'h5' in image[0] and image[0] != 'H5image']
#     #     for image_class in h5image_classes:
#     #         try:
#     #             print 'Testing class ', image_class[0]
#     #             obj = image_class[1](*args, **kwargs)
#     #             # obj.read(obj.filename)
#     #             print 'Success!'
#     #         except Exception:
#     #             continue
#     #         else:
#     #             cls = image_class[1]
#     #             break
#     #     else:
#     #         raise RuntimeError('H5 format not recognized')
#     #     return super(H5image, cls).__new__(cls)
#     #  can call super.__new__ or simply return the instance (obj) and bypass init
#     extensions = ['h5']
#
#     def read(self, filename, frame=None):
#         h5image_classes = [image for image in inspect.getmembers(sys.modules[__name__], inspect.isclass)
#                            if 'H5image' in image[0] and image[0] != 'H5image']
#
#         # this block ensures that the Dxchange class is the last one checked, as it breaks the following loop if
#         # the image does not match this class
#         for image_class in h5image_classes:
#             if 'DXchangeH5image' in str(image_class):
#                 dxchange_class = image_class
#         h5image_classes.remove(dxchange_class)
#         h5image_classes.append(dxchange_class)
#
#         for image_class in h5image_classes:
#             try:
#                 obj = image_class[1](self.data, self.header)
#                 obj.read(filename)
#             except H5ReadError:
#                 # Skip exception and try the next H5 image class
#                 continue
#             else:
#                 # If not error was thrown break out of loop
#                 break
#         else:
#             # If for loop finished without breaking raise ReadError
#             raise H5ReadError('H5 format not recognized')
#         return obj  # return the successfully read object



@register_h5class
class ALS733H5image(fabioimage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['h5']

    def _readheader(self, f):
        fname = f.name  # get filename from file object
        with h5py.File(fname, 'r') as h:
            self.header = dict(h.attrs)

    def read(self, f, frame=None):
        self.readheader(f)

        # Check header for unique attributes
        try:
            if self.header['facility'] != 'als' or self.header['end_station'] != 'bl733':
                raise H5ReadError
        except KeyError:
            raise H5ReadError

        self.filename = f
        if frame is None:
            frame = 0
        return self.getframe(frame)

    @property
    def nframes(self):
        with h5py.File(self.filename, 'r') as h:
            dset = h[h.keys()[0]]
            ddet = dset[dset.keys()[0]]
            if self.isburst:
                frames = sum(map(lambda key: '.edf' in key, ddet.keys()))
            else:
                frames = 1
        return frames

    def __len__(self):
        return self.nframes

    @nframes.setter
    def nframes(self, n):
        pass

    def getframe(self, frame=None):
        if frame is None:
            frame = 0
        f = self.filename
        with h5py.File(f, 'r') as h:
            dset = h[h.keys()[0]]
            ddet = dset[dset.keys()[0]]
            if self.isburst:
                frames = [key for key in ddet.keys() if '.edf' in key]
                dfrm = ddet[frames[frame]]
            elif self.istiled:
                high = ddet[u'high']
                low = ddet[u'low']
                frames = [high[high.keys()[0]], low[low.keys()[0]]]
                dfrm = frames[frame]
            else:
                dfrm = ddet
            self.data = dfrm[0]
        return self.data

    @property
    def isburst(self):
        try:
            with h5py.File(self.filename, 'r') as h:
                dset = h[h.keys()[0]]
                ddet = dset[dset.keys()[0]]
                return not (u'high' in ddet.keys() and u'low' in ddet.keys())
        except AttributeError:
            return False

    @property
    def istiled(self):
        try:
            with h5py.File(self.filename, 'r') as h:
                dset = h[h.keys()[0]]
                ddet = dset[dset.keys()[0]]
                return u'high' in ddet.keys() and u'low' in ddet.keys()
        except AttributeError:
            return False

# @register_h5class
# class GeneralAPSH5image(fabioimage):
#     """
#     Fabio Image class for arbitrary APS H5 structure
#     """
#     DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['h5']
#     def __init__(self, data=None , header=None):
#         super(GeneralAPSH5image, self).__init__(data=data, header=header)
#         self.frames = None
#         self.currentframe = 0
#         self.header = None
#         self._h5 = None
#         self._dgroup = None
#         self.flats = None
#         self.darks = None
#
#     # Context manager for "with" statement compatibility
#     def __enter__(self, *arg, **kwarg):
#         return self
#
#     def __exit__(self, *arg, **kwarg):
#         self.close()
#
#     def _readheader(self, f):
#         #not really useful at this point
#         if self._h5 is not None:
#             self.header=dict(self._h5.attrs)
#             self.header.update(**self._dgroup.attrs)
#
#     def read(self, f, frame=None):
#         self.filename = f
#         if frame is None:
#             frame = 0
#         if self._h5 is None:
#
#             # Check header for unique attributes
#             self._h5 = h5py.File(self.filename, 'r')
#             self._dgroup = self._finddatagroup(self._h5)
#             self.readheader(f)
#
#             self.frames = range(self._dgroup.shape[0])
#
#         dfrm = self._dgroup[self.frames[frame]]
#         self.currentframe = frame
#         self.data = dfrm
#         return self
#
#     def _finddatagroup(self, h5object):
#         keys = h5object.keys()
#         for key in keys:
#             try:
#                 data, data_key = self._check_if_dataset(h5object, key)
#                 break
#             except TypeError:
#                 pass
#
#         try:
#             return data[data_key]
#         except NameError:
#             raise H5ReadError('Unable to find dataset group')
#
#     def _check_if_dataset(self, h5object, key):
#         #recursively find dataset in h5 tree structure
#         if isinstance(h5object[key], h5py.Dataset):
#             return h5object, key
#         else:
#             try:
#                 for lower_key in h5object[key].keys():
#                     return self._check_if_dataset(h5object[key], lower_key)
#             except AttributeError:
#                 pass
#
#     def flatindices(self):
#         nproj = len(self)
#         return [0, nproj - 1]
#
#     @property
#     def nframes(self):
#         return len(self.frames)
#
#     @nframes.setter
#     def nframes(self, n):
#         pass
#
#     def __getitem__(self, item):
#         s = []
#         if not isinstance(item, tuple) and not isinstance(item, list):
#             item = (item, )
#         for n in range(3):
#             if n == 0:
#                 stop = len(self)
#             elif n == 1:
#                 stop = self.data.shape[0]
#             elif n == 2:
#                 stop = self.data.shape[1]
#             if n < len(item) and isinstance(item[n], slice):
#                     start = item[n].start if item[n].start is not None else 0
#                     step = item[n].step if item[n].step is not None else 1
#                     stop = item[n].stop if item[n].stop is not None else stop
#             elif n < len(item) and isinstance(item[n], int):
#                     if item[n] < 0:
#                         start, stop, step = stop + item[n], stop + item[n] + 1, 1
#                     else:
#                         start, stop, step = item[n], item[n] + 1, 1
#             else:
#                 start, step = 0, 1
#
#             s.append((start, stop, step))
#
#         for n, i in enumerate(range(s[0][0], s[0][1], s[0][2])):
#             _arr = self._dgroup[self.frames[i]][slice(*s[1]), slice(*s[2])]
#             if n == 0:  # allocate array
#                 arr = np.empty((len(range(s[0][0], s[0][1], s[0][2])), _arr.shape[0], _arr.shape[1]))
#             arr[n] = _arr
#         if arr.shape[0] == 1:
#             arr = arr[0]
#         return np.squeeze(arr)
#
#     def __len__(self):
#         return self.nframes
#
#     def getframe(self, frame=0):
#         self.data = self._dgroup[self.frames[frame]]
#         return self.data
#
#     def next(self):
#         if self.currentframe < self.__len__() - 1:
#             self.currentframe += 1
#         else:
#             raise StopIteration
#         return self.getframe(self.currentframe)
#
#     def previous(self):
#         if self.currentframe > 0:
#             self.currentframe -= 1
#             return self.getframe(self.currentframe)
#         else:
#             raise StopIteration
#
#     def close(self):
#         self._h5.close()

@register_h5class
class DXchangeH5image(fabioimage):
    """
    Fabio Image class for Data-Exchange HDF5 Datasets
    """
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['h5']
    def __init__(self, data=None, header=None):
        super(DXchangeH5image, self).__init__(data=data, header=header)
        self.currentframe = 0
        self.header = None
        self._h5 = None
        self._dgroup = None
        self._flats = None
        self._darks = None

    # Context manager for "with" statement compatibility
    def __enter__(self, *arg, **kwarg):
        return self

    def __exit__(self, *arg, **kwarg):
        self.close()

    def _readheader(self, f):
        # TODO What data should be read here?
        if self._h5 is not None:
            self.header = {'foo': 'bar'}  # dict(self._h5.attrs)
            # self.header.update(**self._dgroup.attrs)

    def read(self, f, frame=None):
        self.filename = f
        if frame is None:
            frame = 0
        if self._h5 is None:
            self._h5 = h5py.File(self.filename, 'r')
            self._dgroup = self._finddatagroup(self._h5)
        self.readheader(f)
        self.currentframe = frame
        self.nframes = self._dgroup['data'].shape[0]
        self.data = self._dgroup['data'][frame]
        self.frames = range(self._dgroup['data'].shape[0])
        return self

    def _finddatagroup(self, h5object):
        exchange_groups = [key for key in h5object.keys() if 'exchange' in key]
        if len(exchange_groups) > 1:
            raise RuntimeWarning('More than one exchange group found. Will use \'exchange\'\n'
                                 'Need to use logging for this...')
        return h5object[exchange_groups[0]]

    @property
    def flats(self):
        if self._flats is None:
            if 'data_white' in self._dgroup:
                self._flats = self._dgroup['data_white']
        return self._flats

    @property
    def darks(self):
        if self._darks is None:
            if 'data_dark' in self._dgroup:
                self._darks = self._dgroup['data_dark']
        return self._darks

    def __getitem__(self, item):
        return self._dgroup['data'][item]

    def __len__(self):
        return self._dgroup['data'].shape[0]

    def getframe(self, frame=0):
        self.data = self._dgroup['data'][frame]
        return self.data

    def next(self):
        if self.currentframe < self.__len__() - 1:
            self.currentframe += 1
        else:
            raise StopIteration
        return self.getframe(self.currentframe)

    def previous(self):
        if self.currentframe > 0:
            self.currentframe -= 1
            return self.getframe(self.currentframe)
        else:
            raise StopIteration

    def close(self):
        self._h5.close()

    # def flatindices(self):
    #     return None


class TiffStack(object):
    """
    Class for stacking several individual tiffs and viewing as a 3D image in an pyqtgraph ImageView
    """

    def __init__(self, paths, header=None):
        super(TiffStack, self).__init__()
        if isinstance(paths, list):
            self.frames = paths
        elif os.path.isdir(paths):
            self.frames = sorted(glob.glob(os.path.join(paths, '*.tiff')))
        self.currentframe = 0
        self.header= header
        self.rawdata = np.stack([self.getframe(frame) for frame in range(len(self.frames))])

        self._readheader()

    def __len__(self):
        return len(self.frames)

    def getframe(self, frame=0):
        self.data = tifffile.imread(self.frames[frame], memmap=True)
        return self.data

    def _readheader(self):
        #not really useful at this point
        if not self.header:
            self.header = {}
            self.header['shape'] = self.rawdata.shape

    def __getitem__(self, item):
        return self.rawdata[item]

    def close(self):
        pass

class CondensedTiffStack(object):
    """
    Class for 3D tiffs to view in pyqtgraph ImageView - very similar to TiffStack class
    """
    def __init__(self, path, header=None):
        super(CondensedTiffStack, self).__init__()

        self.rawdata = tifffile.imread(path, memmap=True)
        self.frames = range(self.rawdata.shape[0])
        self.header = header
        self._readheader()

    @property
    def classname(self):
        return 'CondensedTiffStack'

    def __len__(self):
        return len(self.frames)

    def _readheader(self):
        #not really useful at this point
        if not self.header:
            self.header = {}
            self.header['shape'] = self.rawdata.shape

    def getframe(self, frame=0):
        return self.rawdata[frame]

    def __getitem__(self, item):
        return self.rawdata[item]

    def close(self):
        pass


class EdfImage(edfimage.EdfImage):
    DEFAULT_EXTENTIONS = DEFAULT_EXTENSIONS = ['edf']

    def read(self, f, frame=None):
        return super(EdfImage, self).read(f, frame)

    def _readheader(self, f):
        super(EdfImage, self)._readheader(f)
        f = f.name.replace('.edf', '.txt')
        if os.path.isfile(f):
            self.header.update(self.scanparas(f))

    @staticmethod
    def scanparas(path):
        if not os.path.isfile(path):
            return dict()

        with open(path, 'r') as f:
            lines = f.readlines()

        paras = OrderedDict()

        # The 7.3.3 txt format is messy, with keyless values, and extra whitespaces

        keylesslines = 0
        for line in lines:
            cells = filter(None, re.split('[=:]+', line))

            key = cells[0].strip()

            if cells.__len__() == 2:
                cells[1] = cells[1].split('/')[0]
                paras[key] = cells[1].strip()
            elif cells.__len__() == 1:
                keylesslines += 1
                paras['Keyless value #' + str(keylesslines)] = key

        return paras


fabioformats._extension_cache = None

class H5ReadError(IOError):
    """
    Exception class raised when checking for the specific schema/structure of an HDF5 file.
    """
    pass


# # Testing
# if __name__ == '__main__':
#     from matplotlib.pyplot import imshow, show
#     data = fabio.open('/home/lbluque/TestDatasetsLocal/dleucopodia.h5') #20160218_133234_Gyroid_inject_LFPonly.h5')
#     # arr = data[-1,:,:] #.getsinogramchunk(slice(0, 512, 1), slice(1000, 1500, 1))
#     slc = (slice(None), slice(None, None, 8), slice(None, None, 8))
#     # arr = data.__getitem__(slc)
#     arr = data.getsinogram(100)
#     # print sorted(data.frames, reverse=True)
#     # print data.darks.shape
#     # print data.flats.shape
#     imshow(arr, cmap='gray')
#     show()
def tests():

    # print(fabio.open('/home/rp/Downloads/Bar5_sample2_th0.098_10.00s_3451.tif').data)
    print(fabio.open('/home/rp/Downloads/MFLP3_50Al_OK_5TH_Sp2_40549-00001(1).fits').data)

if __name__ == '__main__':

    tests()



# --coding: utf-8 --
import numpy as np
from PySide import QtGui, QtCore
import pyqtgraph as pg
from modpkgs import nonesigmod
from collections import OrderedDict

class timelineLinePlot(pg.PlotWidget):
    def __init__(self, frameaxes, xaxes):
        super(timelineLinePlot, self).__init__()
        self.variationcurve = OrderedDict()

    def setData(self, t, **kwargs):
        for key in kwargs.keys():
            if key != 'y' and not self.plotItem.legend:
                self.addLegend()

            if not key in self.variationcurve:
                self.variationcurve[key] = self.plot(name=key, pen=pg.mkPen(
                    color=pg.intColor(kwargs.keys().index(key), len(kwargs.keys()))))
                # self.variationcurve[key].setPen()

            y = kwargs[key]
            l=min(len(t),len(y))
            self.variationcurve[key].setData(x=t[:l], y=y[:l])

    def clear_hack(self):
        self.variationcurve = OrderedDict()
        if self.plotItem.legend:
            self.plotItem.legend.scene().removeItem(self.plotItem.legend)
            self.plotItem.legend = None
        super(timelineLinePlot, self).clear()


class timelineStackPlot(QtGui.QWidget):
    def __init__(self, frameaxes, xaxes):
        super(timelineStackPlot, self).__init__()

        self.slider = QtGui.QSlider(QtCore.Qt.Vertical, self)
        self.slider.setGeometry(500, 500, 30, 10)
        # slider.setFocusPolicy(QtCore.Qt.NoFocus)

        self.plotWidget = pg.PlotWidget(labels={'bottom':(xaxes['label'])})

        layout = QtGui.QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.plotWidget)
        layout.addWidget(self.slider)

        self.slider.valueChanged.connect(self.setOffset)

        self.curves = {}
        self._curvesdata = {}

    def mapSlider(self):
        return (np.exp(self.slider.value() / 100.) - 1) / 10

    def setOffset(self, _):
        for i,curve in enumerate(self.curves.values()):
            x, y = self._curvesdata['x'],self._curvesdata['y']
            curve.setData(x[i], y[i] + i * self.mapSlider())

    def setData(self, t, x, y):
        self._curvesdata={'t':t,'x':x,'y':y}
        for i,t0 in enumerate(t):
            if t0 not in self.curves:
                self.curves[t0] = self.plotWidget.plot(x[i], y[i] + (len(self.curves) * self.mapSlider()))
        self.setColors()
        self.setOffset(0)

    def setColors(self):
        for i,curve in enumerate(self.curves.values()):
            curve.setPen(
                pg.mkPen(color=[(1 - float(i) / len(self.curves)) * 255, float(i) / len(self.curves) * 255, 255]))

    def clear(self):
        self.curves = {}
        self._curvesdata = {}
        self.plotWidget.clear()


class timelineWaterfallPlot(pg.ImageView):
    def __init__(self,frameaxes,xaxes):
        self.axesItem = pg.PlotItem()
        self.axesItem.setLabel('top', frameaxes['label'], units=frameaxes['units'] if 'units' in frameaxes else None)
        self.axesItem.setLabel('left', xaxes['label'], units=xaxes['units'] if 'units' in xaxes else None)
        self.axesItem.axes['left']['item'].setZValue(10)
        self.axesItem.axes['top']['item'].setZValue(10)
        self.axesItem.showAxis('bottom', False)

        kwargs= {'view':self.axesItem}
        super(timelineWaterfallPlot, self).__init__(**kwargs)
        self.view.setAspectLocked(False)

    def setData(self, t, x, y):
        self.setImage(np.array(y))

    def setImage(self, *args, **kwargs):
        kwargs['pos'] = [0, 0]
        super(timelineWaterfallPlot, self).setImage(*args, **kwargs)


class TimelinePlot(QtGui.QTabWidget):
    def __init__(self,frameaxes={'label':'Time','units':'s'},
                      xaxes={'label':u'q (Å⁻¹)'}):
        super(TimelinePlot, self).__init__()
        self.waterfall = timelineWaterfallPlot(frameaxes,xaxes)
        self.lineplot = timelineLinePlot(frameaxes,xaxes)
        self.stackplot = timelineStackPlot(frameaxes,xaxes)
        self.addTab(self.lineplot, 'Line plot')
        self.addTab(self.waterfall, 'Waterfall')
        self.addTab(self.stackplot, 'Stack plot')
        self._data = {'t': []}
        self.setTabPosition(self.West)
        self.currentChanged.connect(self.widgetChanged)

    @QtCore.Slot(object, list, OrderedDict)
    @nonesigmod.pyside_none_deco
    def addData(self, t, *args, **kwargs):
        if len(args) == 1:
            kwargs['y']=args[0]
        elif len(args)==2:
            kwargs['x']=args[0]
            kwargs['y']=args[1]

        if t is None: return

        is1D = True
        if len(kwargs)==2 and 'x' in kwargs and 'y' in kwargs:
            x = kwargs['x']
            y = kwargs['y']
            is1D = False

        if len(self._data['t']) == 0:
            self.setDMode(is1D)

        # append data to currentData
        self._data['t'].append(t)
        for key,value in kwargs.items():
            if not key in self._data: self._data[key] = []
            self._data[key].append(value)

        self.setData()

    plot = addData

    def setData(self):
        self.currentPlot().setData(**self._data)

    def setDMode(self,is1D):
        index = None
        if is1D:
            self.setCurrentIndex(0)
            self.setTabEnabled(0, True)
            self.setTabEnabled(1, False)
            self.setTabEnabled(2, False)
            index = 0  # line plot
        else:
            self.setTabEnabled(0, False)
            self.setTabEnabled(1, True)
            self.setTabEnabled(2, True)
            if self.currentIndex() == 0:
                self.setCurrentIndex(1)
            index = 1  # waterfall plot

        if index is None: index = self.currentIndex()  # otherwise do nothing
        return index

    def guessPlot(self, data):
        # if its a line plot, switch to line plot; if its a 3d plot, switch to a 3d plot or whichever is already selected
        index = None
        if type(data[1]) is tuple:
            self.setTabEnabled(0, False)
            self.setTabEnabled(1, True)
            self.setTabEnabled(2, True)
            if self.currentIndex() == 0:
                self.setCurrentIndex(1)
            index = 1  # waterfall plot

        else:
            self.setCurrentIndex(0)
            self.setTabEnabled(0, True)
            self.setTabEnabled(1, False)
            self.setTabEnabled(2, False)
            index = 0  # line plot

        if index is None: index = self.currentIndex()  # otherwise do nothing
        return index

    def currentPlot(self):
        return self.widget(self.currentIndex())

    def clearData(self):
        self._data = {'t': []}

    def clear(self):
        self.clearData()
        self.currentPlot().clear()
        if hasattr(self.currentPlot(), 'clear_hack'): self.currentPlot().clear_hack()

    def widgetChanged(self, *args, **kwargs):
        self.currentPlot().clear()
        if len(self._data)>1: self.setData()


## Start Qt event loop unless running in interactive mode.
if __name__ == '__main__':
    import sys

    app = QtGui.QApplication([])

    ## Create window with ImageView widget
    win = QtGui.QMainWindow()
    win.resize(800, 800)

    timelineplot = TimelinePlot()
    win.setCentralWidget(timelineplot)
    win.show()
    win.setWindowTitle('TimelinePlot')

    t = 0


    def update():
        global t, timer
        if False:
            data = OrderedDict([('t', t),('FWHM',np.random.rand()),('center',t)])
            t += 1
        elif True:
            data = {'t': t, 'x': np.arange(0, 1, .01), 'y': np.sin(np.arange(0, 1, .01) * t)}
            t += 1
        else:
            data = {'t': t, 'y': np.sin(.01 * t)}
            t += 1
        timelineplot.addData(**data)

        if t == 255: timer.stop()


    timer = pg.QtCore.QTimer()
    timer.timeout.connect(update)
    timer.start(50)

    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()

class XASTimelineWidget(TimelinePlot):
    def __init__(self):
        super(XASTimelineWidget, self).__init__(frameaxes={'label':'Scan'},
                                                xaxes={'label':'Energy','units':'eV'})


def load(path):
    with open(path) as f:
        content = f.read()

    return hig(d)


def parsetodict(content):
    """
    :type content: str
    :param content:
    :return:
    """
    if '=' in content:
        content.partition('=')


#takes string and writes in hig format
def dict2str(d, depth=0):
    content = ''
    for key in d:
        if type(d[key]) is dict:
            content += u'{0}{1} = {{\n'.format(u'\t' * depth, str(key))
            content += dict2str(d[key], depth + 1)
            content = content[:-1] + u'\n{0}}},\n'.format(u'\t' * depth)
        elif type(d[key]) is tuple:
            content += u'{0}{1} = {2},\n'.format(u'\t' * depth, key, '[ ' + ' '.join(map(str, d[key])) + ' ]')
        elif type(d[key]) is list:
            content += u'{0}{1} = {2},\n'.format(u'\t' * depth, key, '[ ' + ' '.join(map(str, d[key])) + ' ]')
        elif type(d[key]) is unicode:
            content += u'{0}{1} = "{2}",\n'.format(u'\t' * depth, key, str(d[key]))
        elif type(d[key]) is str:
            content += u'{0}{1} = "{2}",\n'.format(u'\t' * depth, key, str(d[key]))
        else:
            content += u'{0}{1} = {2},\n'.format(u'\t' * depth, key, str(d[key]))

    content = content[:-2] + u'\n'

    return content


class hig:
    def __init__(self, **d):
        self.__dict__.update(d)

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, item):
        return None

    def __str__(self):
        return dict2str(self.__dict__)

    def write(self, path):
        with open(path,'w') as f:
            f.write(str(self))


if __name__ == '__main__':
    d = {'hipRMCInput': {'instrumentation': {'inputimage': 'data/mysphere.tif',
                                             'imagesize': [512, 512],
                                             'numtiles': 1,
                                             'loadingfactors': [0.111]},
                         'computation': {'runname': "test",
                                         'modelstartsize': [32, 32],
                                         'numstepsfactor': 1000,
                                         'scalefactor': 32}}}
    h = hig(**d)
    print h


# --coding: utf-8 --
import numpy as np
from xicam import config
from PySide import QtCore
import multiprocessing
import time
import pyFAI
import msg

pyFAI_method = 'cython'
from modpkgs import pyFAImod

#
#
# def radialintegrate(dimg, cut=None):
#     print 'WARNING: Histogram binning method of radial integration is in use.'
#
#     center = dimg.experiment.center
#
#     # print(self.config.maskingmat)
#     mask = dimg.experiment.mask
#
#     if mask is None:
#         print("No mask defined, creating temporary empty mask.")
#         mask = np.zeros_like(dimg.data)
#     elif not mask.shape == dimg.data.shape:
#         print("Mask dimensions do not match image dimensions. Mask will be ignored until this is corrected.")
#         mask = np.zeros_like(dimg.data)
#
#     print 'Mask:', mask.shape
#     print 'Image:', dimg.data.shape
#     print 'Center:', dimg.experiment.center
#
#
#
#     # else:
#     #    mask = self.config.maskingmat
#
#     invmask=1-mask
#
#     #mask data
#     data = dimg.data * (invmask)
#
#     print 'invmask:', invmask.shape
#
#     if cut is not None:
#         invmask *= cut
#         data *= cut
#
#     #calculate data radial profile
#     y, x = np.indices(data.shape)
#     r = np.sqrt((x - center[1]) ** 2 + (y - center[0]) ** 2)
#     r = r.astype(np.int)
#
#     tbin = np.bincount(r.ravel(), data.ravel())
#     nr = np.bincount(r.ravel(), invmask.ravel())
#     with np.errstate(divide='ignore', invalid='ignore'):
#         radialprofile = tbin / nr
#
#     q = np.arange(radialprofile.shape[0])
#
#     if dimg.experiment.iscalibrated:
#         # calculate q spacings
#         x = np.arange(radialprofile.shape[0])
#         theta = np.arctan2(x * dimg.experiment.getvalue('Pixel Size X'),
#                            dimg.experiment.getvalue('Detector Distance'))
#         #theta=x*self.config.getfloat('Detector','Pixel Size')*0.000001/self.config.getfloat('Beamline','Detector Distance')
#         wavelength = dimg.experiment.getvalue('Wavelength')
#         q = 4 * np.pi / wavelength * np.sin(theta / 2) * 1e-10
#
#
#         # save integration to file
#         #f = open("integ.csv", "w")
#         #for l, z in zip(q, radialprofile):
#         #    f.write(str(l) + "," + str(z) + "\n")
#         #f.close()
#
#         # remove 0s
#         # (q, radialprofile) = ([qvalue for qvalue, Ivalue in zip(q, radialprofile) if Ivalue > 0],
#         #                      [Ivalue for qvalue, Ivalue in zip(q, radialprofile) if Ivalue > 0])
#         radialprofile = radialprofile * (radialprofile > 0) + 0.0001 * (radialprofile <= 0)
#
#         # import xicam.debug
#         # xicam.debug.showimage(data)
#         # xicam.debug.showimage(invmask)
#         # xicam.debug.showimage(r)
#
#
#     return (q, radialprofile)


def pixel_2Dintegrate(dimg, mask=None):
    centerx = dimg.experiment.getvalue('Center X')
    centery = dimg.experiment.getvalue('Center Y')

    if mask is None:
        msg.logMessage("No mask defined, creating temporary empty mask.",msg.INFO)
        mask = np.zeros_like(dimg.rawdata)

    # mask data
    data = dimg.transformdata * (mask)

    # calculate data radial profile
    x, y = np.indices(data.shape)
    r = np.sqrt((x - centerx) ** 2 + (y - centery) ** 2)
    r = r.astype(np.int)

    tbin = np.bincount(r.ravel(), data.ravel())
    nr = np.bincount(r.ravel(), (mask).ravel())
    radialprofile = tbin / nr

    return radialprofile


def chi_2Dintegrate(imgdata, cen, mu, mask=None, chires=30):
    """
    Integration over r for a chi range. Output is 30*
    """
    if mask is None:
        msg.logMessage("No mask defined, creating temporary empty mask..",msg.INFO)
        mask = np.zeros_like(imgdata)

    # mask data
    data = imgdata * (1 - mask)

    x, y = np.indices(data.shape).astype(np.float)
    r = np.sqrt((x - cen[0]) ** 2 + (y - cen[1]) ** 2)
    r = r.astype(np.int)

    delta = 3

    rinf = mu - delta / 2.
    rsup = mu + delta / 2.

    rmask = ((rinf < r) & (r < rsup) & (x < cen[0])).astype(np.int)
    data *= rmask

    chi = chires * np.arctan((y - cen[1]) / (x - cen[0]))
    chi += chires * np.pi / 2.
    chi = np.round(chi).astype(np.int)
    chi *= (chi > 0)

    tbin = np.bincount(chi.ravel(), data.ravel())
    nr = np.bincount(chi.ravel(), rmask.ravel())
    angleprofile = tbin / nr

    # vimodel = pymodelfit.builtins.GaussianModel()
    # vimodel.mu = np.pi / 2 * 100
    # vimodel.A = np.nanmax(angleprofile)
    # vimodel.fitData(x=np.arange(np.size(angleprofile)), y=angleprofile, weights=angleprofile)
    # vimodel.plot(lower=0, upper=np.pi * 100)
    # print ('len:',len(angleprofile))
    return angleprofile




#@debugtools.timeit
def radialintegratepyFAI(data, mask=None, AIdict=None, cut=None, color=[255, 255, 255], requestkey = None, qvrt = None, qpar = None):
    if mask is None: mask = config.activeExperiment.mask
    if AIdict is None:
        AI = config.activeExperiment.getAI()
        # p1 = AI.get_poni1()
        # p2 = AI.get_poni2()
        # msg.logMessage(('poni:', p1, p2),msg.DEBUG)
    else:
        AI = pyFAI.AzimuthalIntegrator()
        AI.setPyFAI(**AIdict)


    if mask is not None:
        mask = mask.copy()

    msg.logMessage(('image:', data.shape),msg.DEBUG)
    msg.logMessage(('mask:', mask.shape),msg.DEBUG)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None and type(cut) is np.ndarray:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask = mask.astype(bool) & cut.astype(bool)


    (q, radialprofile) = AI.integrate1d(data.T, config.settings['Integration Bins (q)'], mask=1 - mask.T, method=pyFAI_method)  #pyfai uses 0-valid mask

    q = q/10.

    return q.tolist(), radialprofile.tolist(), color, requestkey


def chiintegratepyFAI(data, mask, AIdict, cut=None, color=[255, 255, 255], requestkey = None, qvrt = None, qpar = None, xres=config.settings['Integration Bins (χ)'], yres=config.settings['Integration Bins (q)']):
    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)
    # Always do mask with 1-valid, 0's excluded

    if mask is not None:
        mask = mask.copy()

    msg.logMessage(('image:', data.shape),msg.DEBUG)
    msg.logMessage(('mask:', mask.shape),msg.DEBUG)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask &= cut.astype(bool)
    # data *= cut


    cake, q, chi = AI.integrate2d(data.T, xres, yres, method=pyFAI_method)
    mask, q, chi = AI.integrate2d(mask.T, xres, yres, method=pyFAI_method)

    maskedcake = np.ma.masked_array(cake, mask=mask<=0)

    chiprofile = np.ma.average(maskedcake, axis=1)

    return chi.tolist(), chiprofile.tolist(), color, requestkey

def xintegrate(data, mask, AIdict, cut=None, color=[255, 255, 255], requestkey = None, qvrt = None, qpar = None):
    if mask is not None:
        mask = mask.copy()

    msg.logMessage(('image:', data.shape),msg.DEBUG)
    msg.logMessage(('mask:', mask.shape),msg.DEBUG)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.INFO)

    if cut is not None:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask &= cut.astype(bool)
    # data *= cut


    maskeddata = np.ma.masked_array(data, mask=1-mask)

    xprofile = np.ma.average(maskeddata, axis=1)

    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)
    qx = AI.qArray(data.shape[::-1])[int(AI.getFit2D()['centerY']),:]/10
    qx[:qx.argmin()]*=-1

    return qx.tolist(), xprofile.tolist(), color, requestkey


def zintegrate(data, mask, AIdict, cut=None, color=[255, 255, 255], requestkey = None, qvrt = None, qpar = None):
    if mask is not None:
        mask = mask.copy()

    msg.logMessage(('image:', data.shape),msg.DEBUG)
    msg.logMessage(('mask:', mask.shape),msg.DEBUG)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask &= cut.astype(bool)
    # data *= cut


    maskeddata = np.ma.masked_array(data, mask=1-mask)

    xprofile = np.ma.average(maskeddata, axis=0)

    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)
    qz = AI.qArray(data.shape[::-1])[:,int(AI.getFit2D()['centerX'])]/10
    qz[:qz.argmin()]*=-1

    return qz.tolist(), xprofile.tolist(), color, requestkey


def cake(imgdata, experiment, mask=None,  xres=config.settings['Integration Bins (χ)'], yres=config.settings['Integration Bins (q)']):
    if mask is None:
        mask = np.zeros_like(imgdata)
    AI = experiment.getAI()
    """:type : pyFAI.AzimuthalIntegrator"""

    return AI.integrate2d(imgdata.T, xres, yres, mask=1-mask.T)


def GetArc(Imagedata, center, radius1, radius2, angle1, angle2):
    mask = np.zeros_like(Imagedata)

    centerx = center[0]
    centery = center[1]
    y, x = np.indices(Imagedata.shape)
    r = np.sqrt((x - centerx) ** 2 + (y - centery) ** 2)
    mask = (r > radius1) & (r < radius2)
    theta = np.arctan2(y - centery, x - centerx) / 2 / np.pi * 360
    mask &= (theta > angle1) & (theta < angle2)

    mask = np.flipud(mask)
    return mask * Imagedata

def qintegrate(*args,**kwargs):
    # if dimg.cakemode:
    #     return xintegrate(dimg.cake, cut, callbackcolor)
    # else:
    return radialintegratepyFAI(*args,**kwargs)

def cakexintegrate(data, mask, AIdict, cut=None, color=[255,255,255], requestkey=None, qvrt = None, qpar = None):
    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask &= cut.astype(bool)

    chi = np.arange(-180,180,360/config.settings['Integration Bins (χ)'])

    maskeddata = np.ma.masked_array(data, mask=1-mask)
    xprofile = np.ma.average(maskeddata, axis=1)

    return chi.tolist(), xprofile.tolist(), color, requestkey

def cakezintegrate(data, mask, AIdict, cut=None, color=[255,255,255], requestkey=None, qvrt = None, qpar = None):
    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask &= cut.astype(bool)

    q = np.arange(config.settings['Integration Bins (q)'])*np.max(qpar)/ 10./config.settings['Integration Bins (q)']

    maskeddata = np.ma.masked_array(data, mask=1-mask)
    zprofile = np.ma.average(maskeddata, axis=0)

    return q.tolist(), zprofile.tolist(), color, requestkey


def remeshqintegrate(data, mask, AIdict, cut=None, color=[255, 255, 255], requestkey=None, qvrt = None, qpar = None):

    alphai=config.activeExperiment.getvalue('Incidence Angle (GIXS)')
    msg.logMessage('Incoming angle applied to remeshed q integration: ' + str(alphai),msg.DEBUG)

    if mask is None: mask = config.activeExperiment.mask
    if AIdict is None:
        AI = config.activeExperiment.getAI()
        # p1 = AI.get_poni1()
        # p2 = AI.get_poni2()
        # msg.logMessage(('poni:', p1, p2),msg.DEBUG)
    else:
        AI = pyFAI.AzimuthalIntegrator()
        AI.setPyFAI(**AIdict)


    if mask is not None:
        mask = mask.copy()

    msg.logMessage(('image:', data.shape),msg.DEBUG)
    msg.logMessage(('mask:', mask.shape),msg.DEBUG)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None and type(cut) is np.ndarray:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask = mask.astype(bool) & cut.astype(bool)


    #qpar, qvrt = remesh.remeshqarray(data, None, AI, np.deg2rad(alphai))
    qsquared=qpar**2 + qvrt**2


    remeshcenter=np.unravel_index(qsquared.argmin(),qsquared.shape)

    print 'center?:',remeshcenter

    f2d=AI.getFit2D()
    f2d['centerX']=remeshcenter[0]
    f2d['centerY']=remeshcenter[1]
    AI.setFit2D(**f2d)
    AIdict=AI.getPyFAI()
    msg.logMessage('remesh corrected calibration: '+str(AIdict))

    AI._cached_array["q_center"]=np.sqrt(qsquared).T/10   # This is cheating! pyFAI may give unexpected results here

    (q, radialprofile) = AI.integrate1d(data.T, config.settings['Integration Bins (q)'], mask=1 - mask.T, method=pyFAI_method)  #pyfai uses 0-valid mask

    return q, radialprofile, color, requestkey

def remeshchiintegrate(data,mask,AIdict,cut=None, color=[255,255,255],requestkey=None, qvrt = None, qpar = None):
    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)

    alphai=config.activeExperiment.getvalue('Incidence Angle (GIXS)')
    msg.logMessage('Incoming angle applied to remeshed q integration: ' + str(alphai),msg.DEBUG)

    qsquared=qpar**2 + qvrt**2

    remeshcenter=np.unravel_index(qsquared.argmin(),qsquared.shape)

    f2d=AI.getFit2D()
    f2d['centerX']=remeshcenter[0]
    f2d['centerY']=remeshcenter[1]
    AI.setFit2D(**f2d)
    AIdict=AI.getPyFAI()

    return chiintegratepyFAI(data,mask,AIdict,cut,color,requestkey, qvrt = None, qpar = None)

def remeshxintegrate(data, mask, AIdict, cut=None, color=[255, 255, 255], requestkey=None, qvrt = None, qpar = None):
    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask &= cut.astype(bool)

    center = np.where(np.abs(qpar) == np.abs(qpar).min())
    qx = qpar[:,center[0][0]]/10.

    maskeddata = np.ma.masked_array(data, mask=1 - mask)
    xprofile = np.ma.average(maskeddata, axis=1)

    return qx.tolist(), xprofile.tolist(), color, requestkey

def remeshzintegrate(data, mask, AIdict, cut=None, color=[255, 255, 255], requestkey=None, qvrt = None, qpar = None):
    AI = pyFAI.AzimuthalIntegrator()
    AI.setPyFAI(**AIdict)

    if not mask.shape == data.shape:
        msg.logMessage("No mask match. Mask will be ignored.",msg.WARNING)
        mask = np.ones_like(data)
        msg.logMessage(('emptymask:', mask.shape),msg.DEBUG)

    if cut is not None:
        msg.logMessage(('cut:', cut.shape),msg.DEBUG)
        mask &= cut.astype(bool)

    center = np.where(np.abs(qvrt) == np.abs(qvrt).min())
    qx = -qvrt[center[1][0]]/10.

    maskeddata = np.ma.masked_array(data, mask=1 - mask)
    xprofile = np.ma.average(maskeddata, axis=0)

    return qx.tolist(), xprofile.tolist(), color, requestkey



__author__ = 'rp'

__all__ = ['center_approx', 'integration', 'loader', 'peakfinding', 'remesh', 'saxs_calibration',
           'cosmics', 'variation', 'writer', 'pathtools', 'hig', 'detectors', 'cWarpImage', 'variationoperators']





import numpy as np
from scipy import signal
from scipy.ndimage import filters
from xicam import debugtools
import pyqtgraph as pg
from PySide import QtCore

# maxfiltercoef = 5
# cwtrange = np.arange(1, 100)
#
# maxfiltercoef = 5
# cwtrange = np.arange(3, 100)
# gaussiancentersigma = 2
# gaussianwidthsigma = 5
#
#
# @debugtools.timeit
# def findpeaks(x, y):
#     cwtdata = filters.gaussian_filter1d(
#         filters.gaussian_filter1d(signal.cwt(y, signal.ricker, cwtrange), gaussiancentersigma, axis=1),
#         gaussianwidthsigma, axis=0)
#     maxima = (cwtdata == filters.maximum_filter(cwtdata, 5))
#     maximaloc = np.where(maxima == 1)
#     x = np.array(x)
#     y = np.array(y)
#
#     # print('before',np.array(list(np.array(np.vstack([x[maximaloc[1]], y[maximaloc[1]], maximaloc])))).shape)
#     return list(np.array(np.vstack([x[maximaloc[1]], y[maximaloc[1]], maximaloc])))


wavelet = signal.ricker # wavelet of choice
widths = np.arange(1, 20) # range of widths of the ricker wavelet to search/evaluate
max_distances = widths / 8. # ridgeline connectivity threshold; smaller values gives more peaks; larger values considers overlapping peaks as one
gap_thresh = 4 # threshold number of rows for ridgeline connectivity; smaller values gives more peaks
min_length = 3 # minimum ridgeline length; smaller values gives more peaks
min_snr = 2 # Minimum SNR
noise_perc = 10 # percentile of points below which to consider noise
h = 3 # number of points skipped in finite differences
truncationlow = 10 # low q truncation for zeros
truncationhigh = 50 # high q truncation for zeros


@debugtools.timeit
def findpeaks(x, y):
    peaks = signal.find_peaks_cwt(y,widths,wavelet,max_distances,gap_thresh,min_length,min_snr,noise_perc)
    peaks = peaks[1:]
    return list(np.array(np.vstack([x[peaks],y[peaks],peaks])))


# def findpeaks(x, y, filtersize=(5, 5), gaussianwidthsigma=5, gaussiancentersigma=0, minimumsigma=100, snr=1.2):
# if x is None:
#         x = np.arange(len(y))
#
#     cwtdata = signal.cwt(y, signal.ricker, cwtrange)
#     cwtdata = filters.gaussian_filter1d(cwtdata, gaussiancentersigma, axis=1)
#     cwtdata = filters.gaussian_filter1d(cwtdata, gaussianwidthsigma, axis=0)
#     cwtdata = filters.minimum_filter1d(cwtdata, minimumsigma, axis=0)
#
#     maxima = (cwtdata == filters.maximum_filter(cwtdata, filtersize))
#     maximasigmas, maximaxs = np.where(maxima == 1)
#
#     x = np.array(x)
#     y = np.array(y)
#
#     peakxs = []
#     peaksigmas = []
#
#     plt.plot(maximaxs, y[maximaxs], 'or')
#
#     for mx, sigma in zip(maximaxs, maximasigmas):
#         print mx, sigma
#         print max(0, mx - filtersize[1]), min(len(y), mx + filtersize[1])
#         window = y[max(0, mx - 2 * filtersize[1]):min(len(y), mx + 2 * filtersize[1])]  # maybe scale with m's width?
#         noiselevel = stats.scoreatpercentile(window, 10)
#         print(mx, noiselevel)
#         if y[mx] > snr * noiselevel:
#             peakxs.append(mx)
#             peaksigmas.append(sigma)
#
#     # print maximaloc
#
#     # print('before',np.array(list(np.array(np.vstack([x[maximaloc[1]], y[maximaloc[1]], maximaloc])))).shape)
#     return list(np.array(np.vstack([x[peakxs], y[peakxs], [peakxs, peaksigmas]])))


# TODO: Refactor this class into xicam module so I can get rid of pyside dependency
class peaktooltip:
    def __init__(self, x, y, widget):
        self.q, self.I, self.width, self.index = findpeaks(x, y)
        self.scatterPoints = pg.PlotDataItem(self.q, self.I, size=10, pen=pg.mkPen(None),
                                             symbolPen=None, symbolBrush=pg.mkBrush(255, 255, 255, 120), symbol='o')
        self.display_text = pg.TextItem(text='', color=(176, 23, 31), anchor=(0, 1))
        self.display_text.hide()
        widget.addItem(self.scatterPoints)
        widget.addItem(self.display_text)
        self.scatterPoints.scene().sigMouseMoved.connect(self.onMove)

    def onMove(self, pixelpos):
        # act_pos = self.scatterPoints.mapFromScene(pos)
        #p1 = self.scatterPoints.pointsAt(act_pos)

        # get nearby points
        itempos = self.scatterPoints.mapFromScene(pixelpos)
        itemx = itempos.x()
        itemy = itempos.y()
        pixeldelta = 7
        delta = self.scatterPoints.mapFromScene(QtCore.QPointF(pixeldelta + pixelpos.x(), pixeldelta + pixelpos.y()))
        deltax = delta.x() - itemx
        deltay = -(delta.y() - itemy)
        p1 = [point for point in zip(self.q, self.I) if (itemx - deltax < point[0] and point[0] < itemx + deltax) and (
        itemy - deltay < point[1] and point[1] < itemy + deltay)]
        # print(self.q[1], self.I[1])
        #print(itemx - deltax, itemx + deltax, itemy - deltay, itemy + deltay)
        #print(len(p1))

        """
        :type p1 : pg.graphicsItems.ScatterPlotItem.SpotItem
        """
        if len(p1) != 0:
            self.display_text.setText('q=%f\nI=%f' % (p1[0][0], p1[0][1]))
            self.display_text.setPos(*p1[0])
            self.display_text.show()
        else:
            self.display_text.hide()







__author__ = "Luis Barroso-Luque"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


import numpy as np
import pyqtgraph as pg
import explorer
from PySide import QtGui, QtCore


class DataTreeWidget(QtGui.QTreeWidget):
    """
    Widget for displaying hierarchical python data structures
    (eg, nested dicts, lists, and arrays), adapted from pyqtgraph datatree.
    """

    def __init__(self, parent=None, data=None):
        QtGui.QTreeWidget.__init__(self, parent)
        self.setVerticalScrollMode(self.ScrollPerPixel)
        self.setData(data)
        self.setColumnCount(2)
        self.setHeaderLabels(['Parameter', 'value'])

    def setData(self, data, hideRoot=False):
        """data should be a dictionary."""
        self.clear()
        self.buildTree(data, self.invisibleRootItem(), hideRoot=hideRoot)
        self.expandToDepth(3)
        self.resizeColumnToContents(0)

    def buildTree(self, data, parent, name='', hideRoot=False):
        if hideRoot:
            node = parent
        else:
            node = QtGui.QTreeWidgetItem([name, ""])
            parent.addChild(node)

        if isinstance(data, dict):
            for k in data.keys():
                self.buildTree(data[k], node, str(k))
        elif isinstance(data, list) or isinstance(data, tuple):
            for i in range(len(data)):
                self.buildTree(data[i], node, str(i))
        else:
            node.setText(1, str(data))


class ImageView(pg.ImageView):
    """
    Subclass of PG ImageView to correct z-slider signal behavior, and add coordinate label.
    See pygtgraph.ImageView for documentation
    """
    sigDeletePressed = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        super(ImageView, self).__init__(*args, **kwargs)
        self.scene.sigMouseMoved.connect(self.mouseMoved)

        self.coordsLabel = QtGui.QLabel(' ', parent=self)
        self.coordsLabel.setMinimumHeight(16)
        self.layout().addWidget(self.coordsLabel)
        self.coordsLabel.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignBottom)
        self.setStyleSheet("background-color: rgba(0,0,0,0%)")


    def buildMenu(self):
        super(ImageView, self).buildMenu()
        self.menu.removeAction(self.normAction)

    def keyPressEvent(self, ev):
        super(ImageView, self).keyPressEvent(ev)
        if ev.key() in (QtCore.Qt.Key_Left, QtCore.Qt.Key_Right, QtCore.Qt.Key_Up, QtCore.Qt.Key_Down):
            self.timeLineChanged()
        elif ev.key() == QtCore.Qt.Key_Delete or ev.key() == QtCore.Qt.Key_Backspace:
            self.sigDeletePressed.emit()

    def timeIndex(self, slider):
        ## Return the time and frame index indicated by a slider
        if self.image is None:
            return (0,0)

        t = slider.value()

        xv = self.tVals
        if xv is None:
            ind = int(t)
        else:
            if len(xv) < 2:
                return (0,0)
            totTime = xv[-1] + (xv[-1]-xv[-2])
            inds = np.argwhere(xv <= t)
            if len(inds) < 1:
                return (0,t)
            ind = inds[-1,0]
        return ind, t

    def mouseMoved(self, ev):
        pos = ev
        viewBox = self.imageItem.getViewBox()
        try:
            if viewBox.sceneBoundingRect().contains(pos):
                mousePoint = viewBox.mapSceneToView(pos)
                x, y = map(int, (mousePoint.x(), mousePoint.y()))
                if (0 <= x < self.imageItem.image.shape[0]) & (0 <= y < self.imageItem.image.shape[1]):  # within bounds
                    self.coordsLabel.setText(u"<div style='font-size: 12pt;background-color:#111111;'>x={0},"
                                             u"   <span style=''>y={1}</span>,   <span style=''>I={2}</span>"\
                                             .format(x, y, self.imageItem.image[x, y]))
                else:
                    self.coordsLabel.setText(u"<div style='font-size: 12pt;background-color:#111111;'>x= ,"
                                             u"   <span style=''>y= </span>,   <span style=''>I= </span>")
        except AttributeError:
            pass

class histDialogButton(QtGui.QPushButton):
    """
    Button to connect to a pyqtgraph.HistogramLUTWidget. Sets the maximum and minimum hist value based on
    values provided by user from popup dialog
    """

    def __init__(self, label, *args, **kwargs):
        super(histDialogButton, self).__init__(label, *args, **kwargs)

        self.value = None

    def showDialog(self, min_default, max_default):

        self.dialog = QtGui.QDialog(parent=self.window())
        layout = QtGui.QVBoxLayout()
        text = QtGui.QLabel("Set maximum and minimum values for histogram.")
        layout.addWidget(text)
        layout.addSpacing(5)

        label1 = QtGui.QLabel("Max: ")
        label2 = QtGui.QLabel("Min: ")

        self.field1 = QtGui.QDoubleSpinBox()
        self.field2 = QtGui.QDoubleSpinBox()
        self.field1.setRange(-1000000, 1000000)
        self.field2.setRange(-1000000, 1000000)
        self.field1.setValue(max_default)
        self.field2.setValue(min_default)

        h1 = QtGui.QHBoxLayout()
        h1.addWidget(label1)
        h1.addWidget(self.field1)
        h2 = QtGui.QHBoxLayout()
        h2.addWidget(label2)
        h2.addWidget(self.field2)

        valueButton = QtGui.QPushButton("Ok")
        valueButton.setDefault(True)
        cancelButton = QtGui.QPushButton("Cancel")
        valueButton.clicked.connect(self.return_vals)
        cancelButton.clicked.connect(self.cancel)

        h3 = QtGui.QHBoxLayout()
        h3.addWidget(cancelButton)
        h3.addWidget(valueButton)

        layout.addLayout(h1)
        layout.addLayout(h2)
        layout.addLayout(h3)
        self.dialog.setLayout(layout)
        self.dialog.setWindowTitle("Set histogram bounds")
        self.dialog.exec_()

    def return_vals(self):
        self.value = [self.field2.value(), self.field1.value()]
        self.dialog.accept()


    def cancel(self):
        self.value = None
        self.dialog.reject()

    def connectToHistWidget(self, hist):

        self.hist = hist
        self.clicked.connect(self.setHistValues)

    def setHistValues(self):

        defaults = self.hist.getLevels()
        self.showDialog(defaults[0], defaults[1])

        if self.value:
            self.hist.setLevels(self.value[0], self.value[1])


class dataDialog(QtGui.QDialog):

    """
    Subclass of QDialog to allow for inputs for filenames
    """

    def __init__(self, parent=None, msg=None):
        super(dataDialog, self).__init__(parent)

        # text at top of box
        text = QtGui.QLabel(msg)
        layout = QtGui.QVBoxLayout()
        layout.addWidget(text)
        layout.addSpacing(5)

        # dialog for searching through files
        dialog = explorer.MultipleFileExplorer(self)
        layout.addWidget(dialog)

        horiz_layout = QtGui.QHBoxLayout()
        # horiz_layout.setContentsMargins(0,0,0,0)

        # ok and cancel buttons for user interaction
        button_box = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel)

        # file name viewer
        file_label = QtGui.QLabel("File name: ")
        file_name = QtGui.QLineEdit()
        file_name.setReadOnly(True)
        file_name.setSizePolicy(QtGui.QSizePolicy.MinimumExpanding, QtGui.QSizePolicy.Fixed)

        horiz_layout.addWidget(file_label)
        horiz_layout.addWidget(file_name)
        horiz_layout.addWidget(button_box)

        layout.addSpacing(5)
        layout.addLayout(horiz_layout)
        self.setLayout(layout)
        self.exec_()



class sliceDialog(QtGui.QDialog):

    """
    Subclass of QDialog to allow for more diverse returns from input dialog popups, specifically
    relating to tomography slice previews

    Attributes
    ----------
    field1: int
        integer representing lowest slice number to reconstruct
    field2: int
        integer representing highest slice number to reconstruct
    """

    def __init__(self, parent=None, val1=0, val2=0, maximum=1000):

        super(sliceDialog, self).__init__(parent)

        text = QtGui.QLabel("Which slice(s) would you like to reconstruct?")
        layout = QtGui.QVBoxLayout()
        layout.addWidget(text)

        label1 = QtGui.QLabel("Start slice: ")
        label2 = QtGui.QLabel("End slice: ")
        self.field1 = QtGui.QSpinBox()
        self.field1.setRange(0,maximum)
        self.field1.setValue(val1)
        self.field2 = QtGui.QSpinBox()
        self.field2.setRange(0,maximum)
        self.field2.setValue(val2)


        h1 = QtGui.QHBoxLayout()
        h1.addWidget(label1)
        h1.addWidget(self.field1)
        h2 = QtGui.QHBoxLayout()
        h2.addWidget(label2)
        h2.addWidget(self.field2)

        valueButton = QtGui.QPushButton("Ok")
        valueButton.setDefault(True)
        cancelButton = QtGui.QPushButton("Cancel")
        valueButton.clicked.connect(self.return_vals)
        cancelButton.clicked.connect(self.cancel)
        h3 = QtGui.QHBoxLayout()
        h3.addWidget(cancelButton)
        h3.addWidget(valueButton)

        layout.addLayout(h1)
        layout.addLayout(h2)
        layout.addLayout(h3)
        self.setLayout(layout)
        self.setWindowTitle("Slice Preview")
        self.exec_()

    def return_vals(self):
        try:
            val1 = int(self.field1.text())
            val2 = int(self.field2.text())
            if val1>val2:
                self.value = [val2,val1]
            else:
                self.value = [val1,val2]
        except ValueError:
            self.value = "Slice values must be integers"

        self.accept()


    def cancel(self):
        self.value = None
        self.reject()



#! /usr/bin/env python

import os
from PySide import QtGui
from  xicam.plugins import base
from viewer import ThreeDViewer

import platform
op_sys = platform.system()
# if op_sys == 'Darwin':
#     from Foundation import NSURL

class Viewer3DPlugin(base.plugin):
    name = "3D Viewer"

    def __init__(self, *args, **kwargs):

        self.centerwidget = QtGui.QTabWidget()
        self.centerwidget.currentChanged.connect(self.currentChanged)
        self.centerwidget.setDocumentMode(True)
        self.centerwidget.setTabsClosable(True)
        self.centerwidget.tabCloseRequested.connect(self.tabCloseRequested)

        # DRAG-DROP
        self.centerwidget.setAcceptDrops(True)
        self.centerwidget.dragEnterEvent = self.dragEnterEvent
        self.centerwidget.dropEvent = self.dropEvent
        self.rightwidget = None

        super(Viewer3DPlugin, self).__init__(*args, **kwargs)

    def openfiles(self, paths):
        print paths
        self.activate()
        widget = ThreeDViewer(paths=paths)
        self.centerwidget.addTab(widget, os.path.basename(paths[0]))
        self.centerwidget.setCurrentWidget(widget)

    def dropEvent(self, e):
        for url in e.mimeData().urls():
            if op_sys == 'Darwin':
                fname = str(NSURL.URLWithString_(str(url.toString())).filePathURL().path())
            else:
                fname = str(url.toLocalFile())
            if os.path.isfile(fname):
                self.openfiles([fname])
            e.accept()

    def dragEnterEvent(self, e):
        print(e)
        e.accept()

    def currentChanged(self, index):
        pass

    def tabCloseRequested(self, index):
        self.centerwidget.widget(index).deleteLater()


# -*- coding: UTF-8 -*-
import time
import inspect
import pyqtgraph as pg
import numpy as np
from pipeline import msg

def timeit(method):
    """
    Use this as a decorator to time a function
    """

    def timed(*args, **kw):
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        # print 'Find peaks called from ' + calframe[1][3]
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        # '%r (%r, %r) %2.2f sec'
        #(method.__name__, args, kw, te - ts)
        msg.logMessage('%r  %2.3f sec' % \
              (method.__name__, te - ts),msg.DEBUG)
        return result

    return timed


def frustration():
    msg.logMessage(u"(ﾉಥ益ಥ)ﾉ﻿ ┻━┻",msg.CRITICAL)


def showimage(img):
    image = pg.image(np.fliplr(img))



import pysftp
import paramiko
from globus import GLOBUSError
from spot import SPOTError
from newt import NEWTError
import ssh

__all__ = ['newt', 'spot', 'globus', 'sftp', 'ssh']

# Exceptions raised by clients that we care to handle
EXCEPTIONS = (pysftp.ConnectionException, paramiko.ssh_exception.BadAuthenticationType,
              paramiko.ssh_exception.AuthenticationException, GLOBUSError, SPOTError, GLOBUSError)






__author__ = "Luis Barroso-Luque"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


import numpy as np
import scipy as sp
import pyqtgraph as pg
from PySide import QtCore


class ROImageOverlay(pg.ROI):
    """
    Class for ROI that can be added to an ImageView to overlay an image with the current image in the ImageView
    Currently the difference between the overlap is shown. To change the operations simply override the updateImage
    method


    Attributes
    ----------
    data
        The data being shown in the ImageView where the ROI is to be added
    currentImage : ndarray
        Copy of the current image in the ImageView
    currentIndex : int
        Index of the current image in the ImageView
    flipped : bool
        Specifies if the image overlay is to be flipped


    Parameters
    ----------
    data
        The data being shown in the ImageView where the ROI is to be added
    bg_imageItem
        Image item from image shown in ImageView
    pos : tuple
        Initial position where ROIImageOverlay should be displayed
    constrainX : bool, optional
        Constrains movement of ROI in x direction
    constrainY : bool, optional
        Constrains movement of ROI in y direction
    translateSnap : bool option
        Snap the ROI to image pixels
    kwargs
        Additional keyword arguments accepted by pg.ROI
    """
    sigTranslated = QtCore.Signal(int, int)

    def __init__(self, data, bg_imageItem, pos, constrainX=False, constrainY=True, translateSnap=True, **kwargs):

        size = bg_imageItem.image.shape
        super(ROImageOverlay, self).__init__(pos, translateSnap=translateSnap, size=size, pen=pg.mkPen(None), **kwargs)

        self.data = data
        self.bg_imgeItem = bg_imageItem
        self._y_constrained = constrainY
        self._x_constrained = constrainX
        self._image_overlap = np.empty(size, dtype='float32')
        self._mask = np.zeros(size, dtype=bool)
        self.currentImage = None
        self.currentIndex = None
        self.flipped = False
        self.setCurrentImage(-1)
        self.flipCurrentImage()
        self.imageItem = pg.ImageItem(self.currentImage)
        self.imageItem.setParentItem(self)
        self.updateImage()

    def setCurrentImage(self, idx):
        """
        Set the currentImage attribute to the image in ImageView given by idx

        Parameters
        ----------
        idx : int
            Index of image displayed in ImageView
        """
        self.currentImage = np.array(self.data[idx]).astype('float32')
        self.currentIndex = idx
        if self.flipped:
            self.flipCurrentImage(toggle=False)

    def constrainX(self, val):
        """
        Sets the constraint of ROI movement in x direction
        """
        self._x_constrained = val

    def constrainY(self, val):
        """
        Sets the constraint of ROI movement in y direction
        """
        self._y_constrained = val

    def flipCurrentImage(self, toggle=True):
        """
        Flips the currentImage
        """
        self.currentImage = np.flipud(self.currentImage)
        if toggle:
            self.flipped = not self.flipped

    @property
    def image_overlap(self):
        """
        Returns the overlap array of the ROIImage and the background image. The returned array has the same dimensions
        As bot the ROI image and the background image with values outside the overlap set to zero.
        """
        self._image_overlap.fill(0)
        x, y = self.pos()

        if x == 0:
            x_slc, bg_x_slc = None, None
        elif x < 0:
            x_slc, bg_x_slc = slice(-int(x), None), slice(None, int(x))
        elif x > 0:
            x_slc, bg_x_slc = slice(None, -int(x)), slice(int(x), None)

        if y == 0:
            y_slc, bg_y_slc = None, None
        elif y < 0:
            y_slc, bg_y_slc = slice(-int(y), None), slice(None, int(y))
        elif y > 0:
            y_slc, bg_y_slc = slice(None, -int(y)), slice(int(y), None)

        slc, bg_slc = (x_slc, y_slc), (bg_x_slc, bg_y_slc)
        self._image_overlap[slc] = self.bg_imgeItem.image[bg_slc]
        return self._image_overlap

    def remove_outlier(self, array1, array2, total, thresh = 0.05):
        val = sp.integrate.trapz(array1, array2)
        print 1- (float(val) / total)
        if 1 - (float(val)/total) < thresh:
            return self.remove_outlier(array1[1:-1],array2[1:-1], total, thresh=thresh)
        else:
            return array1, array2

    def updateImage(self, autolevels=False, levels=None):
        """
        Updates the image shown in the ROI to the difference of the current image and the image_overlap
        """
        if levels:
            self.imageItem.setImage(self.currentImage - self.image_overlap, autoLevels=autolevels, levels=levels)
        else:
            self.imageItem.setImage(self.currentImage - self.image_overlap, autoLevels=autolevels)


    def translate(self, *args, **kwargs):
        """
        Override translate method to update the ROI image and emit the current position of the ROI image
        """
        super(ROImageOverlay, self).translate(*args, **kwargs)
        self.updateImage()
        self.sigTranslated.emit(*self.pos())

    def resetImage(self):
        """
        Resets the current image to the current index
        """
        self.setCurrentImage(self.currentIndex)
        self.updateImage()

    def mouseDragEvent(self, ev):
        """
        Override ROI.mouseDragEvent to set all vertical offsets to zero and constrain dragging to horizontal axis
        """
        if ev.isStart():
            if ev.button() == QtCore.Qt.LeftButton:
                self.setSelected(True)
                if self.translatable:
                    self.isMoving = True
                    self.preMoveState = self.getState()
                    self.cursorOffset = self.pos() - self.mapToParent(ev.buttonDownPos())
                    self.sigRegionChangeStarted.emit(self)
                    ev.accept()
                else:
                    ev.ignore()

        elif ev.isFinish():
            if self.translatable:
                if self.isMoving:
                    self.stateChangeFinished()
                self.isMoving = False
            return

        if self.translatable and self.isMoving and ev.buttons() == QtCore.Qt.LeftButton:
            snap = True if (ev.modifiers() & QtCore.Qt.ControlModifier) else None
            newPos = self.mapToParent(ev.pos()) + self.cursorOffset
            if self._y_constrained:
                newPos.y = self.pos().y
            if self._x_constrained:
                newPos.x = self.pos().x
            self.translate(newPos - self.pos(), snap=snap, finish=False)

    def keyPressEvent(self, ev):
        """
        Override the keyPressEvent to have arrow keys move the ROIImageOverlay
        """

        if ev.key() == QtCore.Qt.Key_Right:
            if not self._x_constrained:
                self.translate(pg.Point((1, 0)))
        elif ev.key() == QtCore.Qt.Key_Left:
            if not self._x_constrained:
                self.translate(pg.Point((-1, 0)))
        elif ev.key() == QtCore.Qt.Key_Up:
            if not self._y_constrained:
                self.translate(pg.Point((0, 1)))
        elif ev.key() == QtCore.Qt.Key_Down:
            if not self._y_constrained:
                self.translate(pg.Point((0, -1)))
        ev.accept()


# avg dphi correlation (mask normalized) for each q

import base
from PySide import QtGui
import os

import widgets


class FXSPlugin(base.plugin):
    name = 'FXS'

    def __init__(self, *args, **kwargs):

        self.centerwidget = QtGui.QTabWidget()
        self.centerwidget.currentChanged.connect(self.currentChanged)
        self.centerwidget.setDocumentMode(True)
        self.centerwidget.setTabsClosable(True)
        self.centerwidget.tabCloseRequested.connect(self.tabCloseRequested)

        super(FXSPlugin, self).__init__(*args, **kwargs)

        # self.sigUpdateExperiment.connect()


    def tabCloseRequested(self, index):
        self.centerwidget.widget(index).deleteLater()

    def getCurrentTab(self):
        if self.centerwidget.currentWidget() is None:
            return None
        return self.centerwidget.currentWidget().widget

    def currentChanged(self, index):
        for tab in [self.centerwidget.widget(i) for i in range(self.centerwidget.count())]:
            tab.unload()
        self.centerwidget.currentWidget().load()
        self.imagePropModel.widgetchanged()


    def openfiles(self, paths=None, operation=None, operationname=None):
        self.activate()
        if type(paths) is not list:
            paths = [paths]

        widget = widgets.OOMTabItem(itemclass=widgets.fxsviewer, paths=paths, operation=operation,
                                    operationname=operationname)
        self.centerwidget.addTab(widget, os.path.basename(paths[0]))
        self.centerwidget.setCurrentWidget(widget)

    def currentImage(self):
        return self.getCurrentTab()

    def invalidatecache(self):
        self.getCurrentTab().dimg.invalidatecache()



import os
import time
from functools import partial

import numpy as np
import yaml

import customwidgets
import display
import featuremanager
import ui
from modpkgs.collectionsmod import UnsortableOrderedDict
from pipeline import msg
from xicam import clientmanager as cmanager
from xicam import plugins
from xicam.plugins import base
from PySide import QtGui


class HipGISAXSPlugin(base.plugin):
    name = 'HipGISAXS'

    def __init__(self, *args, **kwargs):

        self.sftp_client = dict()


        self.centerwidget, self.rightwidget, self.leftwidget = ui.load()

        # INIT FORMS
        self.computationForm = None
        self._detectorForm = None
        self._scatteringForm = None


        # SETUP FEATURES
        featuremanager.layout = ui.leftwidget.featuresList
        featuremanager.load()




        # INIT EXPERIMENT
        self.newExperiment()



        # WIREUP CONTROLS
        self.leftwidget.addFeatureButton.clicked.connect(featuremanager.addLayer)
        self.leftwidget.addSubstrateButton.clicked.connect(featuremanager.addSubstrate)
        self.leftwidget.addParticleButton.clicked.connect(featuremanager.addParticle)
        #self.leftwidget.showComputationButton.clicked.connect(self.showComputation)
        self.leftwidget.showDetectorButton.clicked.connect(self.showDetector)
        self.leftwidget.addParticleButton.setMenu(ui.particlemenu)
        self.leftwidget.save.clicked.connect(self.save)
        self.leftwidget.load.clicked.connect(self.load)

        # RUN MENU
        self.runmenu = QtGui.QMenu()
        self.runmenu.addAction(QtGui.QIcon('xicam/gui/icons_34.png'), "Run Local", self.runLocal)
        self.runmenu.addAction(QtGui.QIcon('xicam/gui/icons_40.png'), "Run Remote", self.runRemote)
        self.runmenu.addAction(QtGui.QIcon('xicam/gui/icons_41.png'), "Run Dask", self.runDask)
        self.leftwidget.run.setMenu(self.runmenu)
        self.leftwidget.run.clicked.connect(self.runLocal)

        # inject loginwidget
        from xicam.widgets import login
        self.loginwidget= login.LoginDialog()
        self.leftwidget.layout().addWidget(self.loginwidget)


        # SETUP DISPLAY
        display.load()
        display.redraw()
        self.centerwidget.addWidget(display.viewWidget)

        super(HipGISAXSPlugin, self).__init__(*args, **kwargs)


    def save(self):
        filename, ok = QtGui.QFileDialog.getSaveFileName(self.centerwidget, 'Save Model', os.curdir, '*.model')

        if filename and ok:
            import yaml
            data = {'computation': self.detectorForm.parameter.saveState(),
                    'items':[{'type':feature.__class__.__name__,
                              'parameter':feature.parameter.saveState(),
                              'structure':feature.structure.parameter.saveState() if hasattr(feature,'structure') else None,
                              'ensemble':feature.ensemble.parameter.saveState() if hasattr(feature,'ensemble') else None,
                              } for feature in featuremanager.features],
                    }
            with open(filename,'w') as f:
                yaml.dump(data, f)

    def load(self):
        filename, ok = QtGui.QFileDialog.getOpenFileName(self.centerwidget, 'Load Model', os.curdir,
                                                         '*.model')
        if filename and ok:
            with open(filename,'r') as f:
                data = yaml.load(f)

            self.detectorForm.parameter.restoreState(data['computation'])
            ui.showForm(ui.blankForm)
            featuremanager.clearFeatures()
            for item in data['items']:
                if item['type']=='particle':
                    feature = customwidgets.particle()
                    feature.parameter.restoreState(item['parameter'])
                    feature.structure.parameter.restoreState(item['structure'])
                    feature.ensemble.parameter.restoreState(item['ensemble'])
                    featuremanager.features.append(feature)
                elif item['type']=='layer':
                    feature = customwidgets.layer()
                    feature.parameter.restoreState(item['parameter'])
                    featuremanager.features.append(feature)
                elif item['type']=='substrate':
                    feature = customwidgets.substrate()
                    feature.parameter.restoreState(item['parameter'])
                    featuremanager.features.append(feature)
            featuremanager.update()


    def newExperiment(self):
        pass
        # featuremanager.clearFeatures()
        # featuremanager.addSubstrate()
        # featuremanager.addLayer()

    def showFeature(self, index):
        self.showForm(index.internalPointer().form)


    def showComputation(self):
        if self.computationForm is None:
            self.computationForm = featuremanager.loadform('xicam/gui/guiComputation.ui')
        self.showForm(self.computationForm)

    def showScattering(self):
        self.showForm(self.scatteringForm.form)

    def showDetector(self):
        self.showForm(self.detectorForm.form)

    def showForm(self, form):
        self.rightwidget.addWidget(form)
        self.rightwidget.setCurrentWidget(form)

    @property
    def detectorForm(self):
        if self._detectorForm is None:
            self._detectorForm = customwidgets.detector()
        return self._detectorForm

    @property
    def scatteringForm(self):
        if self._scatteringForm is None:
            self._scatteringForm = customwidgets.scattering()
        return self._scatteringForm

    def genyaml(self):

        shapes = [feature.toDict() for feature in featuremanager.features if type(feature) is customwidgets.particle]
        layers = [feature.toDict() for feature in featuremanager.features if
                  type(feature) in [customwidgets.layer, customwidgets.substrate]]
        unitcells = [feature.structure.toUnitCellDict() for feature in featuremanager.features if
                     type(feature) is customwidgets.particle]
        structures = [feature.structure.toStructureDict() for feature in featuremanager.features if
                      type(feature) is customwidgets.particle]

        out = {'hipGisaxsInput': UnsortableOrderedDict([('version','0.1'),
                                                        ('shapes', shapes),
                                                        ('unitcells', unitcells),
                                                        ('layers', layers),
                                                        ('structures', structures),
                                                        ('computation', self.detectorForm.toDict())])}
        return out

    def writeyaml(self):
        out = self.genyaml()
        # with open('test.json', 'w') as outfile:
        #     json.dump(out, outfile, indent=4)

        with open(os.path.join(os.path.expanduser('~'),'test.yml'), 'w') as outfile:
            yaml.dump(out, outfile, indent=4)

        msg.logMessage(yaml.dump(out, indent=4))

    def runLocal(self):

        self.writeyaml()

        import subprocess
        import sys
        if sys.platform == 'win32':
            p = subprocess.Popen(["hipgisaxs.exe", os.path.join(os.path.expanduser('~'), 'test.yml')], stdout=subprocess.PIPE)
        else:
            p = subprocess.Popen(["./hipgisaxs", os.path.join(os.path.expanduser('~'),'test.yml')], stdout=subprocess.PIPE)
        stdout,stderr=p.communicate()
        stdout=stdout.replace('\r\r','\r')        # Hack to fix double carriage returns
        out = np.array([np.fromstring(line, sep=' ') for line in stdout.splitlines()])
        #msg.logMessage(stderr.read())
        try:
           plugins.plugins['Viewer'].instance.opendata(out)
        except:
           msg.logMessage("Unable to load data....",msg.ERROR)


        # import os
        #
        # d=os.getcwd()
        # import glob
        # dirs = filter(os.path.isdir, glob.glob(os.path.join(d, "*")))
        # dirs.sort(key=lambda x: os.path.getmtime(x))
        #
        # latestdir=dirs[-1]
        # print 'latestdir',latestdir
        # import glob
        # latestout=glob.glob(os.path.join(latestdir,'*.out'))
        # from xicam import plugins
        # print 'latestout',latestout
        # plugins.plugins['Viewer'].instance.openfiles(latestout)

    def runRemote(self):

        self.writeyaml()

        if len(cmanager.ssh_clients):
            client = cmanager.ssh_clients.values()[0]
            self.loginSuccess(client)
        else:

            add_ssh_callback = lambda client: self.loginSuccess(client)
            login_callback = lambda client: cmanager.add_ssh_client(client.host,
                                                                     client,
                                                                     add_ssh_callback)
            ssh_client = cmanager.ssh_client
            self.loginwidget.loginRequest(partial(cmanager.login, login_callback, ssh_client), True)

    def runDask(self):
        import client.dask_active_executor

        if client.dask_active_executor.active_executor is None:
            #warning message
            return

        ae = client.dask_active_executor.active_executor.executor

        def hipgisaxs_func(yaml_str):
          import subprocess
          import time
          timestamp =time.strftime("%Y.%m.%d.%H.%M.%S")
          filename = os.path.join(os.path.expanduser('~'),"test_remote.yml")

          with open(filename, 'w') as outfile:
             outfile.write(yaml_str)

          a = subprocess.check_output(["srun", "--job-name=hipgisaxs", "--nodes=1", "--ntasks=1", "--ntasks-per-node=1", "--time=00:30:00", "/bin/bash", "/users/course79/rungisaxs.sh", filename])
          return a

        self.writeyaml()

        with open(os.path.join(os.path.expanduser('~'),'test.yml'), 'r') as outfile:
          fx_str = outfile.read()

        msg.showMessage('Submitting job to remote executor...')
        future_tag = ae.submit(hipgisaxs_func, fx_str, pure=False)
        msg.showMessage('Job received. Submitting to queue...')

        import time
        while future_tag.status == "pending":
            #msg.showMessage("Waiting for response from server...",timeout=5)
            time.sleep(1)

        if future_tag.status == "failure":
            msg.showMessage("Execution failed.",timeout=5)
            return

        msg.showMessage("Execution complete. Fetching result...",timeout=5)
        result = future_tag.result()
        out = np.array([np.fromstring(line, sep=' ') for line in result.splitlines()])
        msg.logMessage(("result = ", out),msg.DEBUG)

        plugins.plugins['Viewer'].instance.opendata(out)


    def loginSuccess(self,client):
        self.loginwidget.loginResult(True)
        sftp = client.open_sftp()
        timestamp =time.strftime("%Y.%m.%d.%H.%M.%S")
        sftp.put(os.path.join(os.path.expanduser('~'),'test.yml'),timestamp+'.yml')
        sftp.close()
        stdin,stdout,stderr = client.exec_command('hipgisaxs '+timestamp+'.yml')
        buffer = stdout.read()
        out = np.array([np.fromstring(line,sep=' ') for line in buffer.splitlines()])
        msg.logMessage(stderr.read())
        plugins.plugins['Viewer'].instance.opendata(out)




class mainwindow():
    def __init__(self, app):
        self.app = app









from PySide import QtGui
import sys
import base
from xicam import plugins

# Overload for Py2App
def new_load_qt(api_options):
    from PySide import QtCore, QtGui, QtSvg

    return QtCore, QtGui, QtSvg, 'pyside'


from qtconsole import qt_loaders

qt_loaders.load_qt = new_load_qt

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager

import qdarkstyle
import os



class IPythonPlugin(base.plugin):
    name = 'IPython'

    def __init__(self, *args, **kwargs):
        with open('xicam/gui/style.stylesheet', 'r') as f:
            style = f.read()
        style = (qdarkstyle.load_stylesheet() + style)

        kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel()
        kernel = kernel_manager.kernel
        kernel.gui = 'qt4'
        kernel.shell.push(dict(plugins.plugins))

        kernel_client = kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()


        control = RichJupyterWidget()
        control.kernel_manager = kernel_manager
        control.kernel_client = kernel_client
        control.exit_requested.connect(stop)
        control.style_sheet = style
        control.syntax_style = u'monokai'
        control.set_default_style(colors='Linux')

        self.centerwidget = control

        self.rightwidget = None

        self.featureform = None

        self.bottomwidget = None

        self.leftwidget = None

        self.toolbar = None

        super(IPythonPlugin, self).__init__(*args, **kwargs)





"""
Usage: python setup.py install
       python setup.py bdist_wheel
       python setup.py sdist bdist_egg
       twine upload dist/*
"""

# if __name__ == "__main__":
#     print 'You should run his with pip instead!\n Try this:\n\tpip install .\n...or...\n\tpip install xicam'
#     exit(0)

from setuptools import setup, find_packages
from codecs import open
from os import path
import glob
# from numpy.distutils.core import Extension


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:  # rst?
    long_description = f.read()

# EXT = Extension(name='pipeline.cWarpImage',
#                 sources=['cext/cWarpImage.cc', 'cext/remesh.cc'],
#                 extra_compile_args=['-O3', '-ffast-math'],  # '-fopenmp',, '-I/opt/local/include'
#                 # extra_link_args=['-fopenmp'],
#                 include_dirs=[np.get_include()],
#                 )
setup(
    name='xicam',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.2.25',

    description='A synchrotron data analysis interface',
    long_description=long_description,

    # The project's main homepage.
    url='https://bitbucket.org/ronpandolfi/xicam',

    # Author details
    author='Ronald J Pandolfi',
    author_email='ronpandolfi@lbl.gov',

    # Choose your license
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7'
    ],

    # What does your project relate to?
    keywords='synchrotron analysis x-ray scattering',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    package_dir={'xicam':'xicam','xicamlauncher':'xicamlauncher','pipeline':'pipeline'},

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['scipy', 'Cython', 'pyFAI==0.16.0', 'h5py', 'PySide==1.2.2', 'pyqtgraph', 'QDarkStyle',
                      'Pillow', 'pyfits', 'PyOpenGL', 'PyYAML', 'qtconsole','tifffile','pysftp',
                      'requests','dask','distributed','appdirs','futures','scikit-image','imageio','vispy',
                      'pypaws>=0.8.4','matplotlib', 'astropy'],

    # setup_requires=['numpy', 'cython'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        # 'dev': ['check-manifest'],
        # 'test': ['coverage'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
         'xicam': ['gui/*', 'yaml/*','gui/logos/*'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[#('lib/python2.7/site-packages/gui', glob.glob('gui/*')),
    #            ('lib/python2.7/site-packages/yaml/tomography',glob.glob('yaml/tomography/*'))],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'gui_scripts': [
            'xicam=run_xicam',
        ],
    },

    # ext_modules=[EXT],
    include_package_data=True
)



import os
import sys

from PySide import QtGui, QtCore

if sys.platform == 'win32':
    sys.stdout = open(os.path.join(os.path.expanduser('~'),'out.log'),'w')
    sys.stderr = open(os.path.join(os.path.expanduser('~'),'err.log'),'w')

from splash import SplashScreen

def main():
    sys.path.append(os.path.join(os.getcwd(), 'lib/python2.7/lib-dynload'))

    # set QApplication working dir, so that relative paths resolve properly across different install types
    try:
        d=QtCore.QDir(__file__)
        d.cdUp()
        d.cdUp()
        d.setCurrent(d.path())
        print 'QApp root:',QtCore.QDir().current()
    except NameError:
        print 'Could not set QApp root.' # Hopefully this is run as an executable, and this is unnecessary anyway

    for path in sys.path:
        print 'path:', path
    import xicam  # IMPORTANT! DO NOT REMOVE! Xicam must be loaded early to avoid graphical bugs on mac (?!)
    app=QtGui.QApplication(sys.argv)

    pixmap = QtGui.QPixmap("xicam/gui/splash.gif")
    print 'CWD:', os.getcwd()
    if True:  # Disable to bypass splashscreen for testing on windows
        splash = SplashScreen(pixmap, f=QtCore.Qt.WindowStaysOnTopHint | QtCore.Qt.SplashScreen)
        splash.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        splash.setMask(pixmap.mask())
        splash.show()
        splash.raise_()
        splash.activateWindow()
        app.setActiveWindow(splash)

    else:
        import xicam
        from xicam import xglobals

        xglobals.window = xicam.xicamwindow.MyMainWindow(app)

        xglobals.window.ui.show()
        xglobals.window.ui.raise_()
        xglobals.window.ui.activateWindow()
        app.setActiveWindow(xglobals.window.ui)

    app.processEvents()

    sys.exit(app.exec_())



if __name__ == '__main__':
    main()